from itertools import cycle
from typing import List

from scipy.ndimage import gaussian_filter1d
import numpy as np
from matplotlib import pyplot as plt

import glypy

from .glycan_visual_classification import NGlycanCompositionColorizer, NGlycanCompositionOrderer, GlycanLabelTransformer

from .base import ArtistBase
from ..chromatogram_tree import get_chromatogram, Chromatogram, ChromatogramInterface


def split_charge_states(chromatogram):
    charge_states = chromatogram.charge_states
    versions = {}
    last = chromatogram
    for charge_state in charge_states:
        a, b = last.bisect_charge(charge_state)
        if len(a):
            versions[charge_state] = a
        last = b
    return versions


def label_include_charges(chromatogram, *args, **kwargs):
    return "%s-%r" % (default_label_extractor(chromatogram, **kwargs), tuple(chromatogram.charge_states))


def default_label_extractor(chromatogram, **kwargs):
    if chromatogram.composition:
        return str(chromatogram.composition)
    else:
        return "%0.3f %r" % (chromatogram.neutral_mass, tuple(chromatogram.charge_states))


def binsearch(array: np.ndarray, x: float) -> int:
    lo = 0
    hi = len(array)

    while hi != lo:
        mid = (hi + lo) // 2
        y = array[mid]
        err = y - x
        if abs(err) < 1e-6:
            return mid
        elif (hi - 1) == lo:
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return 0


class ColorCycler(object):
    def __init__(self, colors=None):
        if colors is None:
            colors = ["red", "green", "blue", "yellow", "purple", "grey", "black", "orange"]
        self.color_cycler = cycle(colors)

    def __call__(self, *args, **kwargs):
        return next(self.color_cycler)


class NGlycanChromatogramColorizer(object):
    def __call__(self, chromatogram, default_color="black"):
        if chromatogram.composition is None:
            return default_color
        else:
            try:
                return NGlycanCompositionColorizer(chromatogram.glycan_composition)
            except Exception:
                return default_color


n_glycan_colorizer = NGlycanChromatogramColorizer()


class LabelProducer(object):
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, chromatogram, *args, **kwargs):
        return default_label_extractor(chromatogram)


class NGlycanLabelProducer(LabelProducer):
    """Create a square-brace enclosed tuplet of digits denoting the count
    of of a specific set of glycan composition components common to N-Glycans.

    Relies on :class:`GlycanLabelTransformer`

    """

    def __init__(self, monosaccharides=("HexNAc", "Hex", "Fuc", "NeuAc")):
        super(NGlycanLabelProducer, self).__init__()
        self.monosaccharides = monosaccharides
        self.stub = glypy.GlycanComposition()
        for x in monosaccharides:
            self.stub[x] = -99
        self.label_key = GlycanLabelTransformer([self.stub], NGlycanCompositionOrderer).label_key

    def __call__(self, chromatogram, *args, **kwargs):
        if chromatogram.composition is not None:
            return list(
                GlycanLabelTransformer([chromatogram.glycan_composition, self.stub], NGlycanCompositionOrderer)
            )[0]
        else:
            return "%0.3f (%s)" % (chromatogram.neutral_mass, ", ".join(map(str, chromatogram.charge_states)))


n_glycan_labeler = NGlycanLabelProducer()


class AbundantLabeler(LabelProducer):
    def __init__(self, labeler, threshold):
        self.labeler = labeler
        self.threshold = threshold

    def __call__(self, chromatogram, *args, **kwargs):
        if chromatogram.total_signal > self.threshold:
            return self.labeler(chromatogram, *args, **kwargs), True
        else:
            return self.labeler(chromatogram, *args, **kwargs), False


class ChromatogramArtist(ArtistBase):
    default_label_function = staticmethod(default_label_extractor)
    include_points: bool = True

    def __init__(
        self,
        chromatograms,
        ax=None,
        colorizer=None,
        label_peaks: bool = True,
        clip_labels: bool = True,
        should_draw_tandem_points: bool = True,
    ):
        if colorizer is None:
            colorizer = ColorCycler()
        if ax is None:
            fig, ax = plt.subplots(1)

        if len(chromatograms) > 0:
            chromatograms = self._resolve_chromatograms_from_argument(chromatograms)
            chromatograms = [get_chromatogram(c) for c in chromatograms]
        else:
            chromatograms = []
        self.max_points = float("inf")
        if chromatograms:
            self.max_points = max([len(c) for c in chromatograms])
        self.chromatograms = chromatograms
        self.minimum_ident_time = float("inf")
        self.maximum_ident_time = 0
        self.maximum_intensity = 0
        self.ax = ax
        self.default_colorizer = colorizer
        self.legend = None
        self.label_peaks = label_peaks
        self.clip_labels = clip_labels
        self.should_draw_tandem_points = should_draw_tandem_points

    def __iter__(self):
        return iter(self.chromatograms)

    def __getitem__(self, i):
        return self.chromatograms[i]

    def __len__(self):
        return len(self.chromatograms)

    def _resolve_chromatograms_from_argument(self, chromatograms: List[ChromatogramInterface]) -> List[Chromatogram]:
        try:
            # if not hasattr(chromatograms[0], "get_chromatogram"):
            if not get_chromatogram(chromatograms[0]):
                chromatograms = [chromatograms]
        except TypeError:
            chromatograms = [chromatograms]
        return chromatograms

    def draw_generic_chromatogram(
        self, label, rt: np.ndarray, heights: np.ndarray, color: str, fill: bool = False, label_font_size: int = 10
    ):
        if fill:
            s = self.ax.fill_between(rt, heights, alpha=0.25, color=color, label=label)

        else:
            s = self.ax.plot(rt, heights, color=color, label=label, alpha=0.5)[0]

        s.set_gid(str(label) + "-area")
        if self.include_points:
            s = self.ax.scatter(rt, heights, color=color, s=1)
            s.set_gid(str(label) + "-points")
        apex = max(heights)
        apex_ind = heights.index(apex)
        rt_apex = rt[apex_ind]

        if label is not None:
            self.ax.text(rt_apex, apex + 1200, label, ha="center", fontsize=label_font_size, clip_on=self.clip_labels)

    def draw_group(
        self,
        label: str,
        rt: np.ndarray,
        heights: np.ndarray,
        color: str,
        label_peak: bool = True,
        chromatogram: Chromatogram = None,
        label_font_size: int = 10,
    ):
        if chromatogram is not None:
            try:
                key = str(chromatogram.id)
            except AttributeError:
                key = str(id(chromatogram))
        else:
            key = str(label)

        s = self.ax.fill_between(rt, heights, alpha=0.25, color=color, label=label)
        s.set_gid(key + "-area")
        if self.include_points:
            s = self.ax.scatter(rt, heights, color=color, s=1)
            s.set_gid(key + "-points")
        apex = max(heights)
        heights = np.array(heights)
        maximum_height_mask = heights > apex * 0.95
        apex_indices = np.where(maximum_height_mask)[0]
        apex_ind = apex_indices[apex_indices.shape[0] // 2]
        rt_apex = rt[apex_ind]

        if label is not None and label_peak:
            self.ax.text(
                rt_apex,
                min(apex * 1.1, apex + 1200),
                label,
                ha="center",
                fontsize=label_font_size,
                clip_on=self.clip_labels,
            )

    def transform_group(self, rt, heights):
        return rt, heights

    def draw_tandem_points(self, rt: np.ndarray, heights: np.ndarray, tandem_solutions: list, color: str):
        xs = []
        ys = []
        for tandem in tandem_solutions:
            try:
                x = tandem.scan_time
            except AttributeError:
                # we're dealing with a reference or a stub
                continue
            i = binsearch(rt, x)
            x0 = rt[i]
            y0 = heights[i]
            y1 = 0
            if x0 < x:
                if i < len(rt) - 1:
                    x1 = rt[i + 1]
                    y1 = heights[i + 1]
                else:
                    x1 = rt[i]
                    y1 = heights[i]
            else:
                x1 = x0
                y1 = y0
                if i > 0:
                    x0 = rt[i - 1]
                    y0 = heights[i - 1]
                else:
                    x0 = rt[i]
                    y0 = heights[i]
            # interpolate the height at the time coordinate
            den = x1 - x0
            if den == 0:
                y = (y1 + y0) / 2
            else:
                y = (y0 * (x1 - x) + y1 * (x - x0)) / (x1 - x0)
            xs.append(x)
            ys.append(y)
        self.ax.scatter(xs, ys, marker="X", s=35, color=color)

    def process_group(self, composition, chromatogram: Chromatogram, label_function=None, **kwargs):
        if label_function is None:
            label_function = self.default_label_function

        color = self.default_colorizer(chromatogram)

        rt, heights = self.transform_group(*chromatogram.as_arrays())

        self.maximum_ident_time = max(max(rt), self.maximum_ident_time)
        self.minimum_ident_time = min(min(rt), self.minimum_ident_time)

        self.maximum_intensity = max(max(heights), self.maximum_intensity)

        label = label_function(chromatogram, rt=rt, heights=heights, peaks=None)
        if isinstance(label, (str, glypy.GlycanComposition)):
            label = label
            label_peak = True
        else:
            label, label_peak = label
        label_peak = label_peak & self.label_peaks

        self.draw_group(label, rt, heights, color, label_peak, chromatogram, **kwargs)

        try:
            tandem_solutions = chromatogram.tandem_solutions
        except AttributeError:
            tandem_solutions = []

        if self.should_draw_tandem_points:
            self.draw_tandem_points(rt, heights, tandem_solutions, color)

    def _interpolate_xticks(self, xlo, xhi):
        self.ax.set_xlim(xlo - 0.01, xhi + 0.01)
        tick_values = np.linspace(xlo, xhi, min(5, self.max_points))
        self.ax.set_xticks(tick_values)
        self.ax.set_xticklabels(["%0.2f" % v for v in tick_values])

    def layout_axes(
        self, legend: bool = True, axis_font_size: int = 18, axis_label_font_size: int = 16, legend_cols: int = 2
    ):
        self._interpolate_xticks(self.minimum_ident_time, self.maximum_ident_time)
        self.ax.set_ylim(0, self.maximum_intensity * 1.25)
        if legend:
            try:
                self.legend = self.ax.legend(bbox_to_anchor=(1.2, 1.0), ncol=legend_cols, fontsize=10)
            except ValueError:
                # matplotlib 2.1.1 bug compares array-like colors using == and expects a
                # scalar boolean, triggering a ValueError. When this happens, we can't
                # render a legend.
                self.legend = None
        self.ax.axes.spines["right"].set_visible(False)
        self.ax.axes.spines["top"].set_visible(False)
        self.ax.yaxis.tick_left()
        self.ax.xaxis.tick_bottom()
        self.ax.set_xlabel("Retention Time", fontsize=axis_label_font_size)
        self.ax.set_ylabel("Relative Abundance", fontsize=axis_label_font_size)
        self.ax.ticklabel_format(axis="y", style="scientific", scilimits=(0, 0))
        [t.set(fontsize=axis_font_size) for t in self.ax.get_xticklabels()]
        [t.set(fontsize=axis_font_size) for t in self.ax.get_yticklabels()]

    def draw(
        self,
        label_function=None,
        legend: bool = True,
        label_font_size: int = 10,
        axis_label_font_size: int = 16,
        axis_font_size: int = 18,
        legend_cols: int = 2,
    ):
        if label_function is None:
            label_function = self.default_label_function
        for chroma in self.chromatograms:
            composition = chroma.composition
            self.process_group(composition, chroma, label_function, label_font_size=label_font_size)
        self.layout_axes(
            legend=legend,
            axis_label_font_size=axis_label_font_size,
            axis_font_size=axis_font_size,
            legend_cols=legend_cols,
        )
        return self


class SmoothingChromatogramArtist(ChromatogramArtist):
    def __init__(
        self,
        chromatograms,
        ax=None,
        colorizer=None,
        smoothing_factor=1.0,
        label_peaks=True,
        clip_labels=True,
        should_draw_tandem_points: bool = True,
    ):
        super(SmoothingChromatogramArtist, self).__init__(
            chromatograms,
            ax=ax,
            colorizer=colorizer,
            label_peaks=label_peaks,
            clip_labels=clip_labels,
            should_draw_tandem_points=should_draw_tandem_points,
        )
        self.smoothing_factor = smoothing_factor

    def transform_group(self, rt, heights):
        heights = gaussian_filter1d(heights, self.smoothing_factor)
        return rt, heights

    def draw_generic_chromatogram(self, label, rt, heights, color, fill=False, label_font_size=10):
        heights = gaussian_filter1d(heights, self.smoothing_factor)
        if fill:
            s = self.ax.fill_between(rt, heights, alpha=0.25, color=color, label=label)

        else:
            s = self.ax.plot(rt, heights, color=color, label=label, alpha=0.5)[0]

        s.set_gid(str(label) + "-area")
        s = self.ax.scatter(rt, heights, color=color, s=1)
        s.set_gid(str(label) + "-points")
        apex = max(heights)
        apex_ind = np.argmax(heights)
        rt_apex = rt[apex_ind]

        if label is not None:
            self.ax.text(rt_apex, apex + 1200, label, ha="center", fontsize=label_font_size, clip_on=self.clip_labels)


class ChargeSeparatingChromatogramArtist(ChromatogramArtist):
    default_label_function = staticmethod(label_include_charges)

    def process_group(self, composition, chroma, label_function=None, **kwargs):
        if label_function is None:
            label_function = self.default_label_function
        charge_state_map = split_charge_states(chroma)
        for charge_state, component in charge_state_map.items():
            super(ChargeSeparatingChromatogramArtist, self).process_group(
                composition, component, label_function=label_function, **kwargs
            )


class ChargeSeparatingSmoothingChromatogramArtist(ChargeSeparatingChromatogramArtist, SmoothingChromatogramArtist):
    pass


def mass_shift_separating_chromatogram(chroma, ax=None, **kwargs):
    mass_shifts = list(chroma.mass_shifts)
    labels = {}
    for mass_shift in mass_shifts:
        with_mass_shift, _ = chroma.bisect_mass_shift(mass_shift)
        if len(with_mass_shift):
            labels[mass_shift] = with_mass_shift
    mass_shift_plot = SmoothingChromatogramArtist(list(labels.values()), colorizer=lambda *a, **k: "green", ax=ax).draw(
        label_function=lambda *a, **k: tuple(a[0].mass_shifts)[0].name, legend=False, **kwargs
    )
    rt, intens = chroma.as_arrays()
    mass_shift_plot.draw_generic_chromatogram("Total", rt, intens, color="steelblue")
    ymin = mass_shift_plot.ax.get_ylim()[0]
    mass_shift_plot.ax.set_ylim(ymin, intens.max() * 1.01)
    mass_shift_plot.ax.set_title("Mass Shift-Separated\nExtracted Ion Chromatogram", fontsize=20)
    return mass_shift_plot.ax



from typing import Dict, List, Optional, Sequence, Set, Tuple, Union
from dataclasses import dataclass, field

import numpy as np

from glycresoft.chromatogram_tree.mass_shift import MassShift, Unmodified
from glycresoft.structure.denovo import Path
from glycresoft.structure.scan import ScanStub
from glycresoft.structure.structure_loader import FragmentCachingGlycopeptide

from ms_deisotope import isotopic_shift
from ms_deisotope.data_source import ProcessedScan

from glycresoft.tandem.oxonium_ions import (
    OxoniumIndex,
    SignatureSpecification,
    gscore_scanner,
    single_signatures,
    compound_signatures
)

from glycresoft.database.mass_collection import NeutralMassDatabase, MassObject

from ..core_search import GlycanCombinationRecord, IndexedGlycanFilteringPeptideMassEstimator, CoreMotifFinder

from .search_space import PeptideMassFilterBase


@dataclass
class PeptideMassNode:
    peptide_mass: float
    mass_shift: float
    delta_mass: float
    scan_id: str
    glycan_score: float


@dataclass(frozen=True)
class PeptideEdge:
    structure: FragmentCachingGlycopeptide
    nodes: Tuple['SpectrumNode', 'SpectrumNode'] = field(hash=False)
    similarity: float = field(compare=False, hash=False, default=0.0)
    coverage: float = field(compare=False, hash=False, default=0.0)


@dataclass
class SpectrumNode:
    scan: Union[ProcessedScan, ScanStub]
    precursor_mass: float
    peptide_masses: NeutralMassDatabase[PeptideMassNode]
    signatures: Dict[SignatureSpecification, float]
    oxonium_score: float

    @property
    def scan_id(self):
        return self.scan.id

    def load(self):
        if isinstance(self.scan, ScanStub):
            self.scan = self.scan.convert()
        return self

    def unload(self):
        if isinstance(self.scan, ProcessedScan):
            self.scan = ScanStub(self.scan.precursor_information, self.scan.source)
        return self

    @property
    def deconvoluted_peak_set(self):
        return self.load().deconvoluted_peak_set

    def overlaps(self, other: 'SpectrumNode') -> List[Tuple[PeptideMassNode, PeptideMassNode]]:
        return intersecting_peptide_masses(self.peptide_masses, other.peptide_masses)


@dataclass
class IdentifiedSpectrumNode(SpectrumNode):
    structure: FragmentCachingGlycopeptide


# Glycan averagose mass
AVERAGE_MONOMER_MASS: float = 185.5677185226898


def score_path(path: Path, delta_mass: float) -> float:
    n = delta_mass / AVERAGE_MONOMER_MASS
    k = 2.0
    d = np.log(n) * n / k
    return np.log10([p.intensity for p in path.peaks]).sum() / d


def merge_peptide_nodes_database_denovo(database_nodes: List[PeptideMassNode],
                                        denovo_nodes: List[PeptideMassNode], error_tolerance: float=5e-6) -> List[PeptideMassNode]:
    accumulator = []

    db_i = 0
    denovo_i = 0
    db_n = len(database_nodes)
    denovo_n = len(denovo_nodes)

    while db_i < db_n and denovo_i < denovo_n:
        db_node = database_nodes[db_i]
        denovo_node = denovo_nodes[denovo_i]
        if abs(db_node.peptide_mass - denovo_node.peptide_mass) / denovo_node.peptide_mass <= error_tolerance:
            accumulator.append(db_node)
            db_i += 1
            denovo_i += 1
        elif db_node.peptide_mass < denovo_node.peptide_mass:
            accumulator.append(db_node)
            db_i += 1
        else:
            accumulator.append(denovo_node)
            denovo_i += 1

    accumulator.extend(denovo_nodes[denovo_i:])
    accumulator.extend(database_nodes[db_i:])
    return accumulator


def intersecting_peptide_masses(query_nodes: Sequence[PeptideMassNode],
                                reference_nodes: Sequence[PeptideMassNode],
                                error_tolerance: float=5e-6) -> List[Tuple[PeptideMassNode, PeptideMassNode]]:
    accumulator = []

    query_i = 0
    reference_i = 0
    query_n = len(query_nodes)
    reference_n = len(reference_nodes)

    checkpoint = None

    while query_i < query_n and reference_i < reference_n:
        query_node = query_nodes[query_i]
        reference_node = reference_nodes[reference_i]
        if abs(query_node.peptide_mass - reference_node.peptide_mass) / reference_node.peptide_mass <= error_tolerance:
            accumulator.append((query_node, reference_node))
            if checkpoint is None:
                checkpoint = reference_i
            reference_i += 1
        elif query_node.peptide_mass < reference_node.peptide_mass:
            if checkpoint is not None and query_i < (query_n - 1) and abs(query_node.peptide_mass - query_nodes[query_i + 1].peptide_mass) / query_node.peptide_mass < error_tolerance:
                reference_i = checkpoint
                checkpoint = None
            query_i += 1
        else:
            reference_i += 1
    return accumulator


class PeptideGraphBuilder(PeptideMassFilterBase):
    motif_finder: CoreMotifFinder
    signatures: Dict[SignatureSpecification, float]

    def __init__(self, glycan_compositions, product_error_tolerance=0.00002, glycan_score_threshold=0.1,
                 min_fragments=2, peptide_masses_per_scan=100, probing_range_for_missing_precursors=3,
                 trust_precursor_fits=True, fragment_weight=0.56, core_weight=1.42, oxonium_ion_index=None,
                 signature_ion_index=None, oxonium_ion_threshold: float = 0.05, signatures: List[SignatureSpecification]=None):
        super().__init__(glycan_compositions, product_error_tolerance, glycan_score_threshold, min_fragments,
                         peptide_masses_per_scan, probing_range_for_missing_precursors, trust_precursor_fits,
                         fragment_weight, core_weight, oxonium_ion_index, signature_ion_index, oxonium_ion_threshold)
        if signatures is None:
            signatures = signature_ion_index.signatures if signature_ion_index is not None else list(
                single_signatures.keys())
        self.motif_finder = CoreMotifFinder(self.monosaccharides, self.product_error_tolerance)
        self.signatures = signatures

    def collect_denovo_peptide_nodes(self, scan: ProcessedScan, mass_shifts: Optional[List[MassShift]]=None):
        if mass_shifts is None:
            mass_shifts = [Unmodified]
        if not scan.precursor_information.defaulted and self.trust_precursor_fits:
            probe = 0
        else:
            probe = self.probing_range_for_missing_precursors
        precursor_mass = scan.precursor_information.neutral_mass
        nodes = []
        for i in range(probe + 1):
            neutron_shift = self.neutron_shift * i
            for mass_shift in mass_shifts:
                mass_shift_mass = mass_shift.mass
                intact_mass = precursor_mass - mass_shift_mass - neutron_shift
                for (peptide_mass, paths) in self.motif_finder.estimate_peptide_mass(
                            scan,
                            topn=self.peptide_masses_per_scan,
                            mass_shift=mass_shift,
                            simplify=False,
                            query_mass=precursor_mass - neutron_shift,
                        ):

                    if len(paths[0].peaks) < (self.min_fragments + 1):
                        continue

                    delta_mass = intact_mass - peptide_mass
                    nodes.append(
                        PeptideMassNode(peptide_mass, mass_shift, delta_mass,
                                        scan.id, score_path(paths[0], delta_mass))
                    )
        nodes.sort(key=lambda x: x.peptide_mass)
        return nodes

    def collect_database_peptide_nodes(self, scan: ProcessedScan, mass_shifts: Optional[List[MassShift]]=None):
        if mass_shifts is None:
            mass_shifts = [Unmodified]
        if not scan.precursor_information.defaulted and self.trust_precursor_fits:
            probe = 0
        else:
            probe = self.probing_range_for_missing_precursors
        precursor_mass = scan.precursor_information.neutral_mass
        nodes = []
        for i in range(probe + 1):
            neutron_shift = self.neutron_shift * i
            for mass_shift in mass_shifts:
                mass_shift_mass = mass_shift.mass
                intact_mass = precursor_mass - mass_shift_mass - neutron_shift
                for peptide_mass_pred in self.peptide_mass_predictor.estimate_peptide_mass(
                            scan,
                            topn=self.peptide_masses_per_scan,
                            mass_shift=mass_shift,
                            threshold=self.glycan_score_threshold,
                            min_fragments=self.min_fragments,
                            simplify=False,
                            query_mass=precursor_mass - neutron_shift,
                        ):
                    peptide_mass = peptide_mass_pred.peptide_mass
                    nodes.append(
                        PeptideMassNode(peptide_mass, mass_shift, intact_mass - peptide_mass, scan.id, peptide_mass_pred.score)
                    )
        nodes.sort(key=lambda x: x.peptide_mass)
        return nodes

    def handle_scan(self, scan: ProcessedScan, mass_shifts: Optional[List[MassShift]] = None):
        if mass_shifts is None:
            mass_shifts = [Unmodified]
        nodes = self.collect_database_peptide_nodes(scan, mass_shifts)
        nodes.sort(key=lambda x: x.peptide_mass)
        denovo_nodes = self.collect_denovo_peptide_nodes(scan, mass_shifts)
        denovo_nodes.sort(key=lambda x: x.peptide_mass)
        nodes = merge_peptide_nodes_database_denovo(
            nodes, denovo_nodes, self.product_error_tolerance)
        signature_intensities = {}
        if len(scan.deconvoluted_peak_set) > 0:
            base_peak_intensity: float = scan.base_peak.deconvoluted().intensity
        else:
            base_peak_intensity: float = 1.0
        for sig in self.signatures:
            peak = sig.peak_of(scan.deconvoluted_peak_set, self.product_error_tolerance)
            if peak is not None:
                signature_intensities[sig] = peak.intensity / base_peak_intensity
            else:
                signature_intensities[sig] = 0.0
        oxonium_score = gscore_scanner(scan.deconvoluted_peak_set)
        return SpectrumNode(
            scan,
            scan.precursor_information.neutral_mass,
            NeutralMassDatabase(nodes, lambda x: x.peptide_mass),
            signature_intensities,
            oxonium_score
        )


_PeptideMassNode = PeptideMassNode
_intersecting_peptide_masses = intersecting_peptide_masses

from glycresoft._c.tandem.peptide_graph import PeptideMassNode, intersecting_peptide_masses



import re
import logging
from lxml.etree import LxmlError
from pyteomics import mzid
from pyteomics.xml import (
    unitint, unitfloat, unitstr)

try:
    from pyteomics.xml import cvstr
except ImportError:
    def cvstr(x, *a, **kw):
        return str(x)


logger = logging.getLogger("mzid")

MzIdentML = mzid.MzIdentML
_local_name = mzid.xml._local_name
peptide_evidence_ref = re.compile(r"(?P<evidence_id>PEPTIDEEVIDENCE_PEPTIDE_\d+_DBSEQUENCE_)(?P<parent_accession>.+)")


class MultipleProteinMatchesException(Exception):

    def __init__(self, message, evidence_id, db_sequences, key):
        Exception.__init__(self, message)
        self.evidence_id = evidence_id
        self.db_sequences = db_sequences
        self.key = key


class MultipleProteinInfoDict(dict):
    multi = None


class MissingProteinException(Exception):
    pass


class MissingPeptideEvidenceHandler(object):
    parser = re.compile(
        r"(?P<evidence_id>PEPTIDEEVIDENCE_(?P<peptide_id>PEPTIDE_\d+)_(?P<parent_accession>DBSEQUENCE_.+))")

    def __init__(self, full_id):
        self.full_id = full_id

    def __repr__(self):
        return "MissingPeptideEvidenceHandler(%r)" % (self.full_id,)

    def reconstruct_evidence(self):
        match = self.parser.search(self.full_id)
        if match is None:
            raise ValueError("Cannot interpret %r" % (self.full_id,))
        data = match.groupdict()
        result = {
            "isDecoy": False,
            "end": None,
            "start": None,
            "peptide_ref": data["peptide_id"],
            "dBSequence_ref": data['parent_accession'],
            'pre': None,
            "post": None,
            "id": self.full_id
        }
        return result

    @classmethod
    def recover(cls, full_id):
        inst = cls(full_id)
        return inst.reconstruct_evidence()


class Parser(MzIdentML):

    _index_tags = MzIdentML._indexed_tags - {'SearchDatabase', }

    def _handle_ref(self, info, key, value):
        try:
            referenced = self.get_by_id(value, retrieve_refs=True)
        except AttributeError:
            if key == "peptideEvidence_ref":
                referenced = MissingPeptideEvidenceHandler.recover(value)
            else:
                raise AttributeError(key)
        info.update(referenced)
        del info[key]
        info.pop('id', None)

    def _retrieve_refs(self, info, **kwargs):
        multi = None
        for k, v in dict(info).items():
            if k.endswith('_ref'):
                is_multi_db_sequence = peptide_evidence_ref.match(info[k])
                if is_multi_db_sequence and ':' in is_multi_db_sequence.groupdict()['parent_accession']:
                    groups = is_multi_db_sequence.groupdict()
                    evidence_id = groups['evidence_id']
                    db_sequences = groups['parent_accession'].split(':')
                    if len(db_sequences) > 1:
                        multi = MultipleProteinMatchesException(
                            "", evidence_id, db_sequences, k)
                else:
                    try:
                        self._handle_ref(info, k, v)
                    except (KeyError, LxmlError):
                        info['skip'] = True
        info = MultipleProteinInfoDict(info)
        info.multi = multi
        return info, multi

    def _handle_param(self, element, **kwargs):
        """Unpacks cvParam and userParam tags into key-value pairs"""
        types = {'int': unitint, 'float': unitfloat, 'string': unitstr}
        attribs = element.attrib
        unit_info = None
        unit_accesssion = None
        if 'unitCvRef' in attribs or 'unitName' in attribs:
            unit_accesssion = attribs.get('unitAccession')
            unit_name = attribs.get("unitName", unit_accesssion)
            unit_info = unit_name
        accession = attribs.get("accession")
        # value = attribs.get("value", "")
        if 'value' in attribs and attribs['value'] != '':
            try:
                if attribs.get('type') in types:
                    value = types[attribs['type']](attribs['value'], unit_info)
                else:
                    value = unitfloat(attribs['value'], unit_info)
            except ValueError:
                value = unitstr(attribs['value'], unit_info)
            return {cvstr(attribs['name'], accession, unit_accesssion): value}
        else:
            return {'name': cvstr(attribs['name'], accession, unit_accesssion)}

    def _find_by_id_reset(self, *a, **kw):
        return MzIdentML._find_by_id_reset(self, *a, **kw)

    def _insert_param(self, info, param, **kwargs):
        newinfo = self._handle_param(param, **kwargs)
        if not ('name' in info and 'name' in newinfo):
            info.update(newinfo)
        else:
            if not isinstance(info['name'], list):
                info['name'] = [info['name']]
            info['name'].append(newinfo.pop('name'))

    def _find_immediate_params(self, element, **kwargs):
        return element.xpath('./*[local-name()="{}" or local-name()="{}"]'.format("cvParam", "userParam"))

    def _recursive_populate(self, element, info, kwargs):
        for child in element.iterchildren():
            cname = _local_name(child)
            if cname in ('cvParam', 'userParam'):
                self._insert_param(info, child, **kwargs)
            else:
                if cname not in self.schema_info['lists']:
                    info[cname] = self._get_info_smart(child, **kwargs)
                else:
                    info.setdefault(cname, []).append(
                        self._get_info_smart(child, **kwargs))

    def _convert_values(self, element, info, kwargs):
        converters = self._converters
        for k, v in info.items():
            try:
                for t, a in converters.items():
                    if (_local_name(element), k) in self.schema_info[t]:
                        info[k] = a(v)
            except KeyError:
                continue

    def _populate_references(self, element, info, kwargs):
        info = MultipleProteinInfoDict(info)
        infos = [info]
        # resolve refs
        if kwargs.get('retrieve_refs'):
            try:
                _, multi = self._retrieve_refs(info, **kwargs)
            except ValueError as e:
                print(info, e, kwargs)
                raise e
            if multi is not None:
                info.multi = multi
            if info.multi:
                e = info.multi
                if e is not None:
                    evidence_id = e.evidence_id
                    db_sequences = e.db_sequences
                    key = e.key
                    infos = []
                    for name in db_sequences:
                        dup = info.copy()
                        dup[key] = evidence_id + name
                        self._retrieve_refs(dup, **kwargs)
                        infos.append(dup)
        return infos

    def _get_info(self, element, **kwargs):
        """Extract info from element's attributes, possibly recursive.
        <cvParam> and <userParam> elements are treated in a special way."""
        name = _local_name(element)
        if name in {'cvParam', 'userParam'}:
            return self._handle_param(element)

        info = dict(element.attrib)
        # process subelements
        if kwargs.get('recursive'):
            self._recursive_populate(element, info, kwargs)
        else:
            for param in self._find_immediate_params(element):
                self._insert_param(info, param, **kwargs)

        # process element text
        if element.text and element.text.strip():
            stext = element.text.strip()
            if stext:
                if info:
                    info[name] = stext
                else:
                    return stext

        self._convert_values(element, info, kwargs)

        infos = self._populate_references(element, info, kwargs)

        # flatten the excessive nesting
        for info in infos:
            for k, v in dict(info).items():
                if k in self._structures_to_flatten:
                    info.update(v)
                    del info[k]

            # another simplification
            for k, v in dict(info).items():
                if isinstance(v, dict) and 'name' in v and len(v) == 1:
                    info[k] = v['name']
        out = []
        for info in infos:
            if len(info) == 2 and 'name' in info and (
                    'value' in info or 'values' in info):
                name = info.pop('name')
                info = {name: info.popitem()[1]}
            out.append(info)
        if len(out) == 1:
            out = out[0]
        return out



import os
import re
import traceback

from typing import Mapping, TypeVar, Type, Union


import pickle

from functools import partial

import click

from sqlalchemy.exc import OperationalError, ArgumentError
from sqlalchemy.engine.url import _parse_rfc1738_args as parse_engine_uri

from glycresoft.serialize import (
    DatabaseBoundOperation, GlycanHypothesis, GlycopeptideHypothesis,
    SampleRun, Analysis, AnalysisTypeEnum)

from glycresoft.database.builder.glycan import (
    GlycanCompositionHypothesisMerger,
    named_reductions, named_derivatizations)

from glycresoft.database.builder.glycan.synthesis import synthesis_register
from glycresoft.database.builder.glycopeptide.proteomics import mzid_proteome
from glycresoft.chromatogram_tree import (
    MassShift, Formate, Ammonium,
    Sodium, Potassium)

from glycresoft.tandem.glycopeptide.scoring import (
    binomial_score, simple_score, coverage_weighted_binomial, intensity_scorer,
    SpectrumMatcherBase)

from glycresoft.models import ms1_model_features

from glycopeptidepy.utils.collectiontools import decoratordict
from glycopeptidepy.structure.modification import ModificationTable

from glypy import Substituent, Composition

glycan_source_validators = decoratordict()

T = TypeVar('T')


class ModificationValidator(object):
    """Determines whether a provided command line argument can be mapped to a
    valid peptide modification in a default contructed
    :class:`glycopeptidepy.structure.modification.ModificationTable`

    Attributes
    ----------
    table : glycopeptidepy.structure.modification.ModificationTable
        The modification database to look up modifications by name.
    """
    def __init__(self):
        self.table = ModificationTable()

    def validate(self, modification_string):
        try:
            return self.table[modification_string]
        except KeyError:
            return False


class GlycanSourceValidatorBase(DatabaseBoundOperation):
    def __init__(self, database_connection, source, source_type, source_identifier=None):
        DatabaseBoundOperation.__init__(self, database_connection)
        self.source = source
        self.source_type = source_type
        self.source_identifier = source_identifier

    def validate(self):
        raise NotImplementedError()


@glycan_source_validators('text')
@glycan_source_validators('combinatorial')
class TextGlycanSourceValidator(GlycanSourceValidatorBase):
    def __init__(self, database_connection, source, source_type, source_identifier=None):
        super(TextGlycanSourceValidator, self).__init__(
            database_connection, source, source_type, source_identifier)

    def validate(self):
        return os.path.exists(self.source)


@glycan_source_validators("hypothesis")
class HypothesisGlycanSourceValidator(GlycanSourceValidatorBase):
    def __init__(self, database_connection, source, source_type, source_identifier=None):
        super(HypothesisGlycanSourceValidator, self).__init__(
            database_connection, source, source_type, source_identifier)
        self.handle = DatabaseBoundOperation(source)

    def validate(self):
        if self.source_identifier is None:
            click.secho("No value passed through --glycan-source-identifier.", fg='magenta')
            return False
        try:
            hypothesis_id = int(self.source_identifier)
            inst = self.handle.query(GlycanHypothesis).get(hypothesis_id)
            return inst is not None
        except TypeError:
            hypothesis_name = self.source
            inst = self.handle.query(GlycanHypothesis).filter(GlycanHypothesis.name == hypothesis_name).first()
            return inst is not None


@glycan_source_validators("analysis")
class GlycanAnalysisGlycanSourceValidator(GlycanSourceValidatorBase):
    def __init__(self, database_connection, source, source_type, source_identifier=None):
        super(GlycanAnalysisGlycanSourceValidator, self).__init__(
            database_connection, source, source_type, source_identifier)
        self.handle = DatabaseBoundOperation(source)

    def validate(self):
        if self.source_identifier is None:
            click.secho("No value passed through --glycan-source-identifier.", fg='magenta')
            return False
        try:
            analysis_id = int(self.source_identifier)
            inst = self.handle.query(Analysis).filter(
                Analysis.id == analysis_id,
                Analysis.analysis_type == AnalysisTypeEnum.glycan_lc_ms.name)
            return inst is not None
        except TypeError:
            hypothesis_name = self.source
            inst = self.handle.query(Analysis).filter(
                Analysis.name == hypothesis_name,
                Analysis.analysis_type == AnalysisTypeEnum.glycan_lc_ms.name).first()
            return inst is not None


class GlycanHypothesisCopier(GlycanCompositionHypothesisMerger):
    pass


def validate_glycan_source(context, database_connection, source, source_type, source_identifier=None):
    glycan_source_validator_type = glycan_source_validators[source_type]
    glycan_validator = glycan_source_validator_type(database_connection, source, source_type, source_identifier)
    if not glycan_validator.validate():
        click.secho("Could not validate glycan source %s of type %s" % (
            source, source_type), fg='yellow')
        raise click.Abort()


def validate_modifications(context, modifications):
    mod_validator = ModificationValidator()
    for mod in modifications:
        if not mod_validator.validate(mod):
            click.secho("Invalid modification '%s'" % mod, fg='yellow')
            raise click.Abort()


def validate_unique_name(context, database_connection, name, klass):
    handle = DatabaseBoundOperation(database_connection)
    obj = handle.query(klass).filter(
        klass.name == name).first()
    if obj is not None:
        return klass.make_unique_name(handle.session, name)
    else:
        return name


validate_glycopeptide_hypothesis_name = partial(validate_unique_name, klass=GlycopeptideHypothesis)
validate_glycan_hypothesis_name = partial(validate_unique_name, klass=GlycanHypothesis)
validate_sample_run_name = partial(validate_unique_name, klass=SampleRun)
validate_analysis_name = partial(validate_unique_name, klass=Analysis)


def _resolve_protein_name_list(context, args):
    result = []
    for arg in args:
        if isinstance(arg, str):
            if os.path.exists(arg) and os.path.isfile(arg):
                with open(arg) as fh:
                    for line in fh:
                        cleaned = line.strip()
                        if cleaned:
                            result.append(cleaned)
            else:
                result.append(arg)
        else:
            if isinstance(arg, (list, tuple)):
                result.extend(arg)
            else:
                result.append(arg)
    return result


def validate_mzid_proteins(context, mzid_file, target_proteins=tuple(), target_proteins_re=tuple()):
    all_proteins = set(mzid_proteome.protein_names(mzid_file))
    accepted_target_proteins = set()
    target_proteins = _resolve_protein_name_list(context, target_proteins)
    target_proteins_re = _resolve_protein_name_list(context, target_proteins_re)
    for prot in target_proteins:
        if prot in all_proteins:
            accepted_target_proteins.add(prot)
        else:
            click.secho("Could not find protein '%s'" % prot, fg='yellow')
    for prot_re in target_proteins_re:
        pat = re.compile(prot_re)
        hits = 0
        for prot in all_proteins:
            if pat.search(prot):
                accepted_target_proteins.add(prot)
                hits += 1

        if hits == 0:
            click.secho("Pattern '%s' did not match any proteins" % prot_re, fg='yellow')
    if len(accepted_target_proteins) == 0:
        click.secho("Using all proteins", fg='yellow')
    return list(accepted_target_proteins)


def validate_reduction(context, reduction_string):
    if reduction_string is None:
        return None
    try:
        if str(reduction_string).lower() in named_reductions:
            return named_reductions[str(reduction_string).lower()]
        else:
            if len(Composition(str(reduction_string))) > 0:
                return str(reduction_string)
            else:
                raise Exception("Invalid")
    except Exception:
        click.secho("Could not validate reduction '%s'" % reduction_string)
        raise click.Abort("Could not validate reduction '%s'" % reduction_string)


def validate_derivatization(context, derivatization_string):
    if derivatization_string is None:
        return derivatization_string
    if derivatization_string in named_derivatizations:
        return named_derivatizations[derivatization_string]
    subst = Substituent(derivatization_string)
    if len(subst.composition) == 0:
        click.secho("Could not validate derivatization '%s'" % derivatization_string)
        raise click.Abort("Could not validate derivatization '%s'" % derivatization_string)
    else:
        return derivatization_string


class SubstituentParamType(click.types.StringParamType):
    name = "SUBSTITUENT"

    def convert(self, value, param, ctx):
        t = Substituent(value)
        if not t.composition:
            raise ValueError("%s is not a recognized substituent" % value)
        return t


mass_shifts = {
    "ammonium": Ammonium,
    "formate": Formate,
    "sodium": Sodium,
    "potassium": Potassium,
}


def validate_mass_shift(mass_shift_string, multiplicity=1):
    multiplicity = int(multiplicity)
    if mass_shift_string.lower() in mass_shifts:
        return (mass_shifts[mass_shift_string.lower()], multiplicity)
    else:
        try:
            mass_shift_string = str(mass_shift_string)
            composition = Composition(mass_shift_string)
            shift = MassShift(mass_shift_string, composition)
            return (shift, multiplicity)
        except Exception as e:
            click.secho("%r" % (e,))
            click.secho("Could not validate mass_shift %r" % (mass_shift_string,), fg='yellow')
            raise click.Abort("Could not validate mass_shift %r" % (mass_shift_string,))


glycopeptide_tandem_scoring_functions = {
    "binomial": binomial_score.BinomialSpectrumMatcher,
    "simple": simple_score.SimpleCoverageScorer,
    "coverage_weighted_binomial": coverage_weighted_binomial.CoverageWeightedBinomialScorer,
    "log_intensity": intensity_scorer.LogIntensityScorer,
    "penalized_log_intensty": intensity_scorer.FullSignaturePenalizedLogIntensityScorer,
    "peptide_only_cw_binomial": coverage_weighted_binomial.StubIgnoringCoverageWeightedBinomialScorer,
    "log_intensity_v3": (intensity_scorer.LogIntensityScorerReweighted, {}),
    "log_intensity_reweighted": (intensity_scorer.LogIntensityScorerReweighted, {}),
}


def validate_glycopeptide_tandem_scoring_function(context, name):
    if isinstance(name, SpectrumMatcherBase):
        return name, {}
    try:
        scorer = glycopeptide_tandem_scoring_functions[name]
        if not isinstance(scorer, (tuple, list)):
            return scorer, {}
        return scorer
    except KeyError:
        # Custom scorer loading
        if name.startswith("model://"):
            path = name[8:]
            if not os.path.exists(path):
                raise click.Abort("Model file at \"%s\" does not exist." % (path, ))
            with open(path, 'rb') as fh:
                scorer = pickle.load(fh)
                if not isinstance(scorer, (tuple, list)):
                    return scorer, {}
                else:
                    if len(scorer) == 2 and isinstance(scorer[1], Mapping):
                        return scorer
                    else:
                        raise TypeError(
                            "Unpickled scorer contained a tuple, but the second value was not a dictionary!")
        else:
            raise click.Abort("Could not recognize scoring function by name %r" % (name,))


class ChoiceOrURI(click.Choice):
    def get_metavar(self, param):
        return "[%s|model://path]" % ('|'.join(self.choices), )

    def get_missing_message(self, param):
        choice_str = ",\n\t".join(self.choices)
        return "Choose from:\n\t%s\nor specify a model://path" % choice_str

    def convert(self, value, param, ctx):
        # Match through normalization and case sensitivity
        # first do token_normalize_func, then lowercase
        # preserve original `value` to produce an accurate message in
        # `self.fail`
        normed_value = value
        normed_choices = {choice: choice for choice in self.choices}

        if ctx is not None and ctx.token_normalize_func is not None:
            normed_value = ctx.token_normalize_func(value)
            normed_choices = {
                ctx.token_normalize_func(normed_choice): original
                for normed_choice, original in normed_choices.items()
            }

        if not self.case_sensitive:
            normed_value = normed_value.casefold()
            normed_choices = {
                normed_choice.casefold(): original
                for normed_choice, original in normed_choices.items()
            }

        if normed_value in normed_choices:
            return normed_choices[normed_value]

        if value.startswith("model://"):
            return value

        self.fail(
            "invalid value: {value}. (choose from {choices} or specify a \"model://path\")".format(
                value=value, choices=', '.join(self.choices)),
            param,
            ctx,
        )


def get_by_name_or_id(session, model_type: Type[T], name_or_id: Union[str, int]) -> T:
    try:
        object_id = int(name_or_id)
        inst = session.query(model_type).get(object_id)
        if inst is None:
            raise ValueError("No instance of type %s with id %r" %
                             (model_type, name_or_id))
        return inst
    except ValueError:
        try:
            inst = session.query(model_type).filter(
                model_type.name == name_or_id).one()
            return inst
        except Exception:
            raise click.BadParameter("Could not locate an instance of %r with identifier %r" % (
                model_type.__name__, name_or_id))


def validate_database_unlocked(database_connection):
    try:
        db = DatabaseBoundOperation(database_connection)
        db.session.add(GlycanHypothesis(name="_____not_real_do_not_use______"))
        db.session.rollback()
        return True
    except OperationalError:
        return False


def validate_ms1_feature_name(feature_name):
    try:
        return ms1_model_features[feature_name]
    except KeyError:
        raise click.Abort(
            "Could not recognize scoring feature by name %r" % (
                feature_name,))


def strip_site_root(type, value, tb):
    msg = traceback.format_exception(type, value, tb)
    sanitized = []
    for i, line in enumerate(msg):
        if 'site-packages' in line:
            sanitized.append(line.split("site-packages")[1])
        else:
            sanitized.append(line)
    print(''.join(sanitized))


class RelativeMassErrorParam(click.types.FloatParamType):
    name = 'PPM MASS ERR'

    def convert(self, value, param, ctx):
        value = super(RelativeMassErrorParam, self).convert(value, param, ctx)
        if value > 1e-3:
            click.secho(
                f"Translating {value} PPM -> {value / 1e-6}",
                fg='yellow'
            )
            value /= 1e-6
        if value <= 0:
            self.fail("mass error value must be greater than 0.")
        return value


class DatabaseConnectionParam(click.types.StringParamType):
    name = "CONN"

    def __init__(self, exists=False):
        self.exists = exists

    def convert(self, value, param, ctx):
        value = super(DatabaseConnectionParam, self).convert(value, param, ctx)
        try:
            parse_engine_uri(value)
            return value
        except ArgumentError:
            # not a uri
            if self.exists:
                if not os.path.exists(value):
                    raise self.fail(
                        "Database file {} does not exist.".format(value), param, ctx)
            return value


class SmallSampleFDRCorrectionParam(click.ParamType):
    name = '0-1.0/on/off/auto'

    error_message_template = (
        "%s is not a valid option for FDR strategy selection. "
        "If numerical, it must be between 0 and 1.0, otherwise it must "
        "be one of auto, on, or off."
    )

    def convert(self, value, param, ctx):
        value = value.lower()
        try:
            threshold = float(value)
            if threshold > 1 or threshold < 0:
                self.fail(self.error_message_template % value, param, ctx)
        except ValueError:
            value = value.strip()
            if value == 'auto':
                threshold = 0.5
            elif value == 'off':
                threshold = 0.0
            elif value == 'on':
                threshold = 1.0
            else:
                self.fail(self.error_message_template % value, param, ctx)
            return threshold



from glycresoft.structure import KeyTransformingDecoratorDict


def transform_key(key):
    return key.replace("-", "_")


ms1_model_features = KeyTransformingDecoratorDict(transform_key)


def register_feature(name, feature):
    ms1_model_features[name] = feature


def available_features():
    return list(ms1_model_features)


def get_feature(name):
    return ms1_model_features[name]



cimport cython

from ms_deisotope._c.peak_set cimport DeconvolutedPeak, DeconvolutedPeakSet
from glycresoft._c.chromatogram_tree.mass_shift cimport MassShiftBase
from glycresoft._c.structure.fragment_match_map cimport FragmentMatchMap


cdef class SpectrumMatchBase(object):
    cdef:
        public object scan
        public object target
        public MassShiftBase mass_shift



cdef class SpectrumMatcherBase(SpectrumMatchBase):
    cdef:
        public DeconvolutedPeakSet spectrum
        public FragmentMatchMap solution_map
        public dict annotations


cdef class ScoreSet(object):
    cdef:
        public float glycopeptide_score
        public float peptide_score
        public float glycan_score
        public float glycan_coverage
        public float stub_glycopeptide_intensity_utilization
        public float oxonium_ion_intensity_utilization
        public int n_stub_glycopeptide_matches
        public float peptide_coverage
        public float total_signal_utilization


    cpdef bint _lt(self, ScoreSet other)
    cpdef bint _gt(self, ScoreSet other)
    cpdef bint _eq(self, ScoreSet other)
    cpdef bytearray pack(self)

    cpdef list values(self)

    @staticmethod
    cdef ScoreSet _create(float glycopeptide_score, float peptide_score, float glycan_score, float glycan_coverage,
                          float stub_glycopeptide_intensity_utilization, float oxonium_ion_intensity_utilization,
                          int n_stub_glycopeptide_matches, float peptide_coverage, float total_signal_utilization)

cdef class FDRSet(object):
    cdef:
        public double total_q_value
        public double peptide_q_value
        public double glycan_q_value
        public double glycopeptide_q_value

    cpdef bint _lt(self, FDRSet other)
    cpdef bint _gt(self, FDRSet other)
    cpdef bint _eq(self, FDRSet other)
    cpdef bytearray pack(self)

    @staticmethod
    cdef FDRSet _create(double total_q_value, double peptide_q_value, double glycan_q_value, double glycopeptide_q_value)


@cython.final
@cython.freelist(1000)
cdef class PeakFoundRecord(object):
    cdef:
        public DeconvolutedPeak peak
        public bint checked

    @staticmethod
    cdef inline PeakFoundRecord _create(DeconvolutedPeak peak, bint checked):
        cdef PeakFoundRecord self = PeakFoundRecord.__new__(PeakFoundRecord)
        self.peak = peak
        self.checked = checked
        return self


@cython.final
cdef class PeakLabelMap(object):
    cdef:
        public dict name_to_peaks

    @staticmethod
    cdef PeakLabelMap _create()

    cdef inline void add(self, str name, DeconvolutedPeak peak)
    cdef inline DeconvolutedPeak get(self, str name, bint* found)


import logging
import operator

from ms_deisotope.peak_dependency_network.intervals import SpanningMixin

from glycresoft.structure.lru import LRUCache
from glycresoft.database.mass_collection import ConcatenatedDatabase


logger = logging.getLogger("glycresoft.intervals")


class QueryIntervalBase(SpanningMixin):
    def __init__(self, center, start, end):
        self.center = center
        self.start = start
        self.end = end

    def copy(self):
        return QueryIntervalBase(self.center, self.start, self.end)

    def __hash__(self):
        return hash((self.start, self.center, self.end))

    def __eq__(self, other):
        return (self.start, self.center, self.end) == (other.start, other.center, other.end)

    def __repr__(self):
        return "QueryInterval(%0.4f, %0.4f)" % (self.start, self.end)

    def extend(self, other):
        self.start = min(self.start, other.start)
        self.end = max(self.end, other.end)
        self.center = (self.start + self.end) / 2.
        return self

    def scale(self, x):
        new = QueryIntervalBase(
            self.center,
            self.center - ((self.center - self.start) * x),
            self.center + ((self.end - self.center) * x))
        return new

    def difference(self, other):
        if self.start < other.start:
            if self.end < other.start:
                return self.copy()
            else:
                return QueryIntervalBase(self.center, self.start, other.start)
        elif self.start > other.end:
            return self.copy()
        elif self.start <= other.end:
            return QueryIntervalBase(self.center, other.end, self.end)
        else:
            return self.copy()


class PPMQueryInterval(QueryIntervalBase):

    def __init__(self, mass, error_tolerance=2e-5):
        self.center = mass
        self.start = mass - (mass * error_tolerance)
        self.end = mass + (mass * error_tolerance)


class FixedQueryInterval(QueryIntervalBase):

    def __init__(self, mass, width=3):
        self.center = mass
        self.start = mass - width
        self.end = mass + width


try:
    has_c = True
    _QueryIntervalBase = QueryIntervalBase
    _PPMQueryInterval = PPMQueryInterval
    _FixedQueryInterval = FixedQueryInterval

    from glycresoft._c.structure.intervals import (
        QueryIntervalBase, PPMQueryInterval, FixedQueryInterval)
except ImportError:
    has_c = False


class IntervalSet(object):

    def __init__(self, intervals=None):
        if intervals is None:
            intervals = list()
        self.intervals = sorted(intervals, key=lambda x: x.center)
        self._total_count = None
        self.compute_total_count()

    def _invalidate(self):
        self._total_count = None

    @property
    def total_count(self):
        if self._total_count is None:
            self.compute_total_count()
        return self._total_count

    def compute_total_count(self):
        self._total_count = 0
        for interval in self:
            self._total_count += interval.size
        return self._total_count

    def __iter__(self):
        return iter(self.intervals)

    def __len__(self):
        return len(self.intervals)

    def __getitem__(self, i):
        return self.intervals[i]

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.intervals)

    def _repr_pretty_(self, p, cycle):
        return p.pretty(self.intervals)

    def find_insertion_point(self, mass):
        lo = 0
        hi = len(self)
        if hi == 0:
            return 0, False
        while hi != lo:
            mid = (hi + lo) // 2
            x = self[mid]
            err = x.center - mass
            if abs(err) <= 1e-9:
                return mid, True
            elif (hi - lo) == 1:
                return mid, False
            elif err > 0:
                hi = mid
            else:
                lo = mid
        raise ValueError((hi, lo, err, len(self)))

    def extend_interval(self, target, expansion):
        logger.debug("Extending %r by %r", target, expansion)
        self._invalidate()
        target.extend(expansion)
        i, _ = self.find_insertion_point(target.center)
        consolidate = False
        if i > 0:
            before = self[i - 1]
            consolidate |= before.overlaps(target)
            # print(before, target)
        if i < len(self) - 1:
            after = self[i + 1]
            consolidate |= after.overlaps(target)
            # print(after, target)
        if consolidate:
            logger.debug("Consolidation was required for extension of %r by %r", target, expansion)
            self.consolidate()
            new_interval = self.find_interval(target)
            logger.debug("Consolidated interval %r spans the original %r", new_interval, target)
            return new_interval
        return target

    def insert_interval(self, interval):
        center = interval.center
        n = len(self)
        if n != 0:
            index, matched = self.find_insertion_point(center)
            index += 1
            if matched and self[index - 1].overlaps(interval):
                new_group = self.extend_interval(self[index - 1], interval)
                return new_group
            if index < n and interval.overlaps(self[index]):
                new_group = self.extend_interval(self[index], interval)
                return new_group
            if index == 1:
                if self[index - 1].center > center:
                    index -= 1
        else:
            index = 0
        self._insert_interval(index, interval)
        return interval

    def _insert_interval(self, index, interval):
        self._invalidate()
        self.intervals.insert(index, interval)

    def find_interval(self, query):
        lo = 0
        n = hi = len(self)
        while hi != lo:
            mid = (hi + lo) // 2
            x = self[mid]
            err = x.center - query.center
            if err == 0 or x.contains_interval(query):
                return x
            elif (hi - 1) == lo:
                best_err = abs(err)
                best_i = mid
                if mid < (n - 1):
                    err = abs(self[mid + 1].center - query.center)
                    if err < best_err:
                        best_err = err
                        best_i = mid + 1
                if mid > -1:
                    err = abs(self[mid - 1].center - query.center)
                    if err < best_err:
                        best_err = err
                        best_i = mid - 1
                return self[best_i]
            elif err > 0:
                hi = mid
            else:
                lo = mid

    def find(self, mass, ppm_error_tolerance):
        return self.find_interval(PPMQueryInterval(mass, ppm_error_tolerance))

    def remove_interval(self, center):
        self._invalidate()
        ix, match = self.find_insertion_point(center)
        self.intervals.pop(ix)

    def clear(self):
        self.intervals = []

    def consolidate(self):
        intervals = list(self)
        self.clear()
        if len(intervals) == 0:
            return
        result = []
        last = intervals[0]
        for current in intervals[1:]:
            if last.overlaps(current):
                last.extend(current)
            else:
                result.append(last)
                last = current
        result.append(last)
        for r in result:
            self.insert_interval(r)


class LRUIntervalSet(IntervalSet):

    def __init__(self, intervals=None, max_size=1000):
        super(LRUIntervalSet, self).__init__(intervals)
        self.max_size = max_size
        self.current_size = len(self)
        self.lru = LRUCache()
        for item in self:
            self.lru.add_node(item)

    def insert_interval(self, interval):
        if self.current_size == self.max_size:
            self.remove_lru_interval()
        super(LRUIntervalSet, self).insert_interval(interval)

    def _insert_interval(self, index, interval):
        super(LRUIntervalSet, self)._insert_interval(index, interval)
        self.lru.add_node(interval)
        self.current_size += 1

    def extend_interval(self, target, expansion):
        self.lru.remove_node(target)
        try:
            result = super(LRUIntervalSet, self).extend_interval(target, expansion)
            self.lru.add_node(result)
            return result
        except Exception:
            self.lru.add_node(target)
            raise

    def find_interval(self, query):
        match = super(LRUIntervalSet, self).find_interval(query)
        if match is not None:
            try:
                self.lru.hit_node(match)
            except KeyError:
                self.lru.add_node(match)
        return match

    def remove_lru_interval(self):
        lru_interval = self.lru.get_least_recently_used()
        logger.debug("Removing LRU interval %r", lru_interval)
        self.lru.remove_node(lru_interval)
        self.remove_interval(lru_interval.center)
        self.current_size -= 1

    def clear(self):
        super(LRUIntervalSet, self).clear()
        self.lru = LRUCache()
        self.current_size = len(self)

    def consolidate(self):
        super(LRUIntervalSet, self).consolidate()
        self.current_size = len(self)


class MassIntervalNode(SpanningMixin):
    """Contains a NeutralMassDatabase object and provides an interval-like
    API. Intended for use with IntervalSet.

    Attributes
    ----------
    center : float
        The central point of mass for the interval
    end : float
        The upper-most mass in the wrapped collection
    group : NeutralMassDatabase
        The collection of massable objects wrapped
    growth : int
        The number of times the interval had been expanded
    size : int
        The number of items in :attr:`group`
    start : float
        The lower-most mass in the wrapped collection
    """
    def __init__(self, interval):
        self.wrap(interval)
        self.growth = 0

    def __hash__(self):
        return hash((self.start, self.center, self.end))

    def __eq__(self, other):
        return (self.start, self.center, self.end) == (other.start, other.center, other.end)

    def wrap(self, interval):
        """Updates the internal state of the interval to wrap

        Parameters
        ----------
        interval : NeutralMassDatabase
            The new collection to wrap.
        """
        self.group = interval
        try:
            self.start = interval.lowest_mass
            self.end = interval.highest_mass
        except IndexError:
            self.start = 0
            self.end = 0
        self.center = (self.start + self.end) / 2.
        self.size = len(self.group)

    def __repr__(self):
        return "%s(%0.4f, %0.4f, %r)" % (
            self.__class__.__name__,
            self.start, self.end, len(self.group) if self.group is not None else 0)

    def _combine_databases(self, current_group, new_group):
        return current_group.merge(new_group, operator.attrgetter('id'))

    def extend(self, new_data):
        """Add the components of `new_data` to `group` and update
        the interval's internal state

        Parameters
        ----------
        new_data : MassIntervalNode
            Iterable of massable objects
        """
        start = min(self.start, new_data.start)
        end = max(self.end, new_data.end)

        self.wrap(self._combine_databases(self.group, new_data.group))
        self.growth += 1

        # max will deal with 0s correctly
        self.end = max(end, self.end)
        # min will prefer 0 and not behave as expected
        if start != 0 and self.start != 0:
            self.start = min(start, self.start)
        elif start != 0:
            # self.start is 0
            self.start = start
        # otherwise they're both 0 and we are out of luck

    def __iter__(self):
        return iter(self.group)

    def __getitem__(self, i):
        return self.group[i]

    def __len__(self):
        return len(self.group)

    def search_mass(self, *args, **kwargs):
        """A proxy for :meth:`NeutralMassDatabase.search_mass`
        """
        return self.group.search_mass(*args, **kwargs)

    def search_mass_ppm(self, *args, **kwargs):
        return self.group.search_mass_ppm(*args, **kwargs)

    def constrain(self, lower, higher):
        subset = self.group.search_between(lower, higher)
        start = max(lower, self.start)
        end = min(higher, self.end)
        self.wrap(subset)
        self.end = end
        self.start = start
        return self


class ConcatenateMassIntervalNode(MassIntervalNode):

    def _combine_databases(self, current_group, new_group):
        if isinstance(current_group, ConcatenatedDatabase):
            return current_group.add(new_group)
        return ConcatenatedDatabase([current_group, new_group])



import time

from collections import defaultdict
from typing import Dict, FrozenSet, Iterable, List, Tuple, TYPE_CHECKING
from uuid import uuid4

from sqlalchemy.orm import Query, aliased

from ms_deisotope.output.common import (ScanDeserializerBase, ScanBunch)

from glycresoft.task import TaskBase

from .connection import DatabaseBoundOperation
from .spectrum import (
    MSScan, PrecursorInformation, SampleRun, DeconvolutedPeak)

from .analysis import Analysis, _AnalysisParametersProps
from .chromatogram import (
    Chromatogram,
    MassShiftSerializer,
    CompositionGroupSerializer,
    ChromatogramSolution,
    GlycanCompositionChromatogram,
    UnidentifiedChromatogram)

from .hypothesis import (
    GlycanComposition,
    GlycanCombinationGlycanComposition,
    GlycanCombination,
    Glycopeptide)

from .tandem import (
    GlycopeptideSpectrumMatch,
    GlycopeptideSpectrumSolutionSet,
    SpectrumClusterBase,
    GlycopeptideSpectrumCluster,
    GlycanCompositionSpectrumCluster,
    UnidentifiedSpectrumCluster,
    GlycanCompositionChromatogramToGlycanCompositionSpectrumCluster,
    UnidentifiedChromatogramToUnidentifiedSpectrumCluster)

from .identification import (
    AmbiguousGlycopeptideGroup,
    IdentifiedGlycopeptide,
    IdentifiedGlycopeptideSummary)


if TYPE_CHECKING:
    from glycresoft.tandem.glycopeptide.identified_structure import (
        IdentifiedGlycopeptide as MemoryIdentifiedGlycopeptide
    )


class AnalysisSerializer(DatabaseBoundOperation, TaskBase):
    _analysis: Analysis
    _analysis_id: int
    _analysis_name: str
    _seed_analysis_name: str
    _peak_lookup_table: Dict[Tuple[str, DeconvolutedPeak], int]
    _mass_shift_cache: MassShiftSerializer
    _composition_cache: CompositionGroupSerializer
    _node_peak_map: Dict
    _scan_id_map: Dict[str, int]
    _chromatogram_solution_id_map: Dict
    _tandem_cluster_cache: Dict[FrozenSet, SpectrumClusterBase]

    def __init__(self, connection, sample_run_id, analysis_name, analysis_id=None):
        DatabaseBoundOperation.__init__(self, connection)
        session = self.session
        self.sample_run_id = sample_run_id

        self._analysis = None
        self._analysis_id = analysis_id
        self._analysis_name = None
        self._seed_analysis_name = analysis_name
        self._peak_lookup_table = None
        self._mass_shift_cache = MassShiftSerializer(session)
        self._composition_cache = CompositionGroupSerializer(session)
        self._node_peak_map = dict()
        self._scan_id_map = self._build_scan_id_map()
        self._chromatogram_solution_id_map = dict()
        self._tandem_cluster_cache = dict()

    def __repr__(self):
        return "AnalysisSerializer(%s, %d)" % (self.analysis.name, self.analysis_id)

    def set_peak_lookup_table(self, mapping: Dict[Tuple, DeconvolutedPeak]):
        self._peak_lookup_table = mapping

    def build_peak_lookup_table(self, minimum_mass=500):
        peak_loader = DatabaseScanDeserializer(self._original_connection, sample_run_id=self.sample_run_id)
        accumulated = peak_loader.ms1_peaks_above(minimum_mass)
        peak_mapping = {x[:2]: x[2] for x in accumulated}
        self.set_peak_lookup_table(peak_mapping)

    def _build_scan_id_map(self) -> Dict[str, int]:
        return dict(self.session.query(
            MSScan.scan_id, MSScan.id).filter(
                MSScan.sample_run_id == self.sample_run_id))

    @property
    def analysis(self):
        if self._analysis is None:
            self._construct_analysis()
        return self._analysis

    @property
    def analysis_id(self):
        if self._analysis_id is None:
            self._construct_analysis()
        return self._analysis_id

    @property
    def analysis_name(self):
        if self._analysis_name is None:
            self._construct_analysis()
        return self._analysis_name

    def set_analysis_type(self, type_string):
        self.analysis.analysis_type = type_string
        self.session.add(self.analysis)
        self.session.commit()

    def set_parameters(self, parameters: Dict):
        self.analysis.parameters = parameters
        self.session.add(self.analysis)
        self.commit()

    def _retrieve_analysis(self):
        self._analysis = self.session.query(Analysis).get(self._analysis_id)
        self._analysis_name = self._analysis.name

    def _create_analysis(self):
        self._analysis = Analysis(
            name=self._seed_analysis_name, sample_run_id=self.sample_run_id,
            uuid=str(uuid4()))
        self.session.add(self._analysis)
        self.session.flush()
        self._analysis_id = self._analysis.id
        self._analysis_name = self._analysis.name

    def _construct_analysis(self):
        if self._analysis_id is not None:
            self._retrieve_analysis()
        else:
            self._create_analysis()

    def save_chromatogram_solution(self, solution, commit=False):
        result = ChromatogramSolution.serialize(
            solution, self.session, analysis_id=self.analysis_id,
            peak_lookup_table=self._peak_lookup_table,
            mass_shift_cache=self._mass_shift_cache,
            composition_cache=self._composition_cache,
            scan_lookup_table=self._scan_id_map,
            node_peak_map=self._node_peak_map)
        try:
            self._chromatogram_solution_id_map[solution.id] = result.id
        except AttributeError:
            pass
        if commit:
            self.commit()
        return result

    def save_glycan_composition_chromatogram_solution(self, solution, commit=False):
        try:
            cluster = self.save_glycan_composition_spectrum_cluster(solution)
            cluster_id = cluster.id
        except AttributeError:
            cluster_id = None
        result = GlycanCompositionChromatogram.serialize(
            solution, self.session, analysis_id=self.analysis_id,
            peak_lookup_table=self._peak_lookup_table,
            mass_shift_cache=self._mass_shift_cache,
            composition_cache=self._composition_cache,
            scan_lookup_table=self._scan_id_map,
            node_peak_map=self._node_peak_map)
        if cluster_id is not None:
            self.session.execute(
                GlycanCompositionChromatogramToGlycanCompositionSpectrumCluster.insert(), {
                    "chromatogram_id": result.id,
                    "cluster_id": cluster_id
                })
            self.session.flush()
        try:
            self._chromatogram_solution_id_map[solution.id] = result.solution.id
        except AttributeError:
            pass

        if commit:
            self.commit()
        return result

    def save_unidentified_chromatogram_solution(self, solution, commit=False):
        try:
            cluster = self.save_unidentified_spectrum_cluster(solution)
            cluster_id = cluster.id
        except AttributeError:
            cluster_id = None
        result = UnidentifiedChromatogram.serialize(
            solution, self.session, analysis_id=self.analysis_id,
            peak_lookup_table=self._peak_lookup_table,
            mass_shift_cache=self._mass_shift_cache,
            composition_cache=self._composition_cache,
            scan_lookup_table=self._scan_id_map,
            node_peak_map=self._node_peak_map)
        if cluster_id is not None:
            self.session.execute(
                UnidentifiedChromatogramToUnidentifiedSpectrumCluster.insert(), {
                    "chromatogram_id": result.id,
                    "cluster_id": cluster_id
                })
            self.session.flush()
        try:
            self._chromatogram_solution_id_map[solution.id] = result.solution.id
        except AttributeError:
            pass

        if commit:
            self.commit()
        return result

    def save_glycan_composition_spectrum_cluster(self, glycan_spectrum_cluster, commit=False):
        inst = GlycanCompositionSpectrumCluster.serialize(
            glycan_spectrum_cluster,
            self.session,
            self._scan_id_map,
            self._mass_shift_cache,
            self.analysis_id)
        if commit:
            self.commit()
        return inst

    def save_unidentified_spectrum_cluster(self, unidentified_spectrum_cluster, commit=False):
        inst = UnidentifiedSpectrumCluster.serialize(
            unidentified_spectrum_cluster,
            self.session,
            self._scan_id_map,
            self._mass_shift_cache,
            self.analysis_id)
        if commit:
            self.commit()
        return inst

    def save_glycopeptide_identification(self, identification: 'MemoryIdentifiedGlycopeptide',
                                         commit: bool=False):
        _start_time = time.time()
        if identification.chromatogram is not None:
            chromatogram_solution = self.save_chromatogram_solution(
                identification.chromatogram, commit=False)
            chromatogram_solution_id = chromatogram_solution.id
        else:
            chromatogram_solution_id = None
        _chromatogram_saving_elapsed = time.time() - _start_time

        # Compute the cluster key to avoid duplicating the storage of duplicate matches
        cluster_key = GlycopeptideSpectrumCluster.compute_key(identification.tandem_solutions)
        _cluster_key_elapsed = time.time() - (_chromatogram_saving_elapsed + _start_time)
        if cluster_key in self._tandem_cluster_cache:
            cluster = self._tandem_cluster_cache[cluster_key]
            _cluster_serialized_elapsed = 0
        else:
            cluster = GlycopeptideSpectrumCluster.serialize(
                identification, self.session, self._scan_id_map, self._mass_shift_cache,
                analysis_id=self.analysis_id)
            self._tandem_cluster_cache[cluster_key] = cluster
            _cluster_serialized_elapsed = time.time() - (_cluster_key_elapsed +
                                                         _chromatogram_saving_elapsed + _start_time)

        _total_elapsed = (_cluster_serialized_elapsed + _cluster_key_elapsed + _chromatogram_saving_elapsed)

        if _total_elapsed > 5:
            _cluster_size = sum(map(len, identification.tandem_solutions))
            _chromatogram_size = len(identification.chromatogram) if identification.chromatogram else 0
            self.log(f"..... Saving {identification} took {_total_elapsed:0.2f} sec "
                     f"(chromatogram {_chromatogram_saving_elapsed:0.2f}s of {_chromatogram_size} nodes,"
                     f" key {_cluster_key_elapsed:0.2f}s,"
                     f" cluster {_cluster_serialized_elapsed:0.2f}s of {_cluster_size} entries)")

        cluster_id = cluster.id

        best_match = identification.best_spectrum_match

        inst = IdentifiedGlycopeptide.serialize(
            identification, self.session, chromatogram_solution_id, cluster_id, analysis_id=self.analysis_id)

        if identification.chromatogram is not None:
            apex_time = identification.apex_time
            total_signal = identification.total_signal
            start_time = identification.start_time
            end_time = identification.end_time
            weighted_neutral_mass = identification.weighted_neutral_mass

            best_spectrum_match_id = None
            sset: GlycopeptideSpectrumSolutionSet = self.session.query(
                GlycopeptideSpectrumSolutionSet).join(MSScan).filter(
                MSScan.scan_id == best_match.scan_id,
                GlycopeptideSpectrumSolutionSet.analysis_id == self.analysis_id
            ).first()

            if sset is not None:
                db_best_match = sset.spectrum_matches.filter(
                    GlycopeptideSpectrumMatch.structure_id == inst.structure_id
                ).first()
                if best_match is not None:
                    if db_best_match is not None:
                        best_spectrum_match_id = db_best_match.id

            if best_spectrum_match_id is None:
                self.log(f"Failed to resolve best spectrum match for {identification} in the database")

            summary = IdentifiedGlycopeptideSummary(
                id=inst.id,
                weighted_neutral_mass=weighted_neutral_mass,
                apex_time=apex_time,
                total_signal=total_signal,
                start_time=start_time,
                end_time=end_time,
                best_spectrum_match_id=best_spectrum_match_id,
            )
            self.session.add(summary)
            self.session.flush()
        if commit:
            self.commit()
        return inst

    def save_glycopeptide_identification_set(self, identification_set: Iterable['MemoryIdentifiedGlycopeptide'],
                                             commit: bool=False):
        cache = defaultdict(list)
        no_chromatograms = []
        out = []
        n = len(identification_set)
        i = 0

        chromatogram_node_counter = 0
        gpsms_solution_counter = 0
        spectra_counter = 0
        for case in identification_set:
            i += 1
            if i % 100 == 0:
                self.log(
                    f"{i * 100. / n:0.2f}% glycopeptides saved. ({i}/{n}), {case.structure}:"
                    f"MS2={case.ms2_score:0.2f}:{case.q_value:0.3f}, MS1={case.ms1_score:0.2f}:{case.total_signal:0.2e}"
                    f"({spectra_counter} spectra, {gpsms_solution_counter} solutions, "
                    f"{chromatogram_node_counter} XIC points)"
                )
            gpsms_solution_counter += sum(map(len, case.tandem_solutions))
            spectra_counter += len(case.tandem_solutions)
            chromatogram_node_counter += len(case.chromatogram) if case.chromatogram is not None else 0
            saved = self.save_glycopeptide_identification(case)
            if case.chromatogram is not None:
                cache[case.chromatogram].append(saved)
            else:
                no_chromatograms.append(saved)
            out.append(saved)
        for chromatogram, members in cache.items():
            AmbiguousGlycopeptideGroup.serialize(
                members, self.session, analysis_id=self.analysis_id)
        for case in no_chromatograms:
            AmbiguousGlycopeptideGroup.serialize(
                [case], self.session, analysis_id=self.analysis_id)
        if commit:
            self.commit()
        return out

    def commit(self):
        self.session.commit()


class AnalysisDeserializer(DatabaseBoundOperation, _AnalysisParametersProps):
    def __init__(self, connection: str, analysis_name=None, analysis_id=None):
        DatabaseBoundOperation.__init__(self, connection)

        if analysis_name is analysis_id is None:
            analysis_id = 1

        self._analysis = None
        self._analysis_id = analysis_id
        self._analysis_name = analysis_name

    @property
    def name(self):
        return self.analysis_name

    @property
    def chromatogram_scoring_model(self):
        try:
            return self.analysis.parameters["scoring_model"]
        except KeyError:
            from glycresoft.models import GeneralScorer
            return GeneralScorer

    def _retrieve_analysis(self):
        if self._analysis_id is not None:
            self._analysis = self.session.query(Analysis).get(self._analysis_id)
            self._analysis_name = self._analysis.name
        elif self._analysis_name is not None:
            self._analysis = self.session.query(Analysis).filter(Analysis.name == self._analysis_name).one()
            self._analysis_id = self._analysis.id
        else:
            raise ValueError("No Analysis identification information provided")

    @property
    def analysis(self) -> Analysis:
        if self._analysis is None:
            self._retrieve_analysis()
        return self._analysis

    @property
    def analysis_id(self) -> int:
        if self._analysis_id is None:
            self._retrieve_analysis()
        return self._analysis_id

    @property
    def analysis_name(self) -> str:
        if self._analysis_name is None:
            self._retrieve_analysis()
        return self._analysis_name

    @property
    def parameters(self):
        return self.analysis.parameters

    def load_unidentified_chromatograms(self):
        from glycresoft.chromatogram_tree import ChromatogramFilter
        node_type_cache = dict()
        scan_id_cache = dict()
        q = self.query(UnidentifiedChromatogram).filter(
            UnidentifiedChromatogram.analysis_id == self.analysis_id).all()
        chroma = ChromatogramFilter([c.convert(
            chromatogram_scoring_model=self.chromatogram_scoring_model,
            node_type_cache=node_type_cache,
            scan_id_cache=scan_id_cache) for c in q])
        return chroma

    def load_glycan_composition_chromatograms(self):
        from glycresoft.chromatogram_tree import ChromatogramFilter
        node_type_cache = dict()
        scan_id_cache = dict()
        q = self.query(GlycanCompositionChromatogram).filter(
            GlycanCompositionChromatogram.analysis_id == self.analysis_id).all()
        chroma = ChromatogramFilter([c.convert(
            chromatogram_scoring_model=self.chromatogram_scoring_model,
            node_type_cache=node_type_cache,
            scan_id_cache=scan_id_cache) for c in q])
        return chroma

    def load_identified_glycopeptides_for_protein(self, protein_id):
        q = self.query(IdentifiedGlycopeptide).join(Glycopeptide).filter(
            IdentifiedGlycopeptide.analysis_id == self.analysis_id,
            Glycopeptide.protein_id == protein_id)
        gps = [c.convert() for c in q]
        return gps

    def load_identified_glycopeptides(self, min_q_value: float=0.2):
        q = self.query(IdentifiedGlycopeptide).filter(
            IdentifiedGlycopeptide.analysis_id == self.analysis_id).all()
        gps = IdentifiedGlycopeptide.bulk_convert(q, min_q_value=min_q_value)
        return gps

    def load_glycans_from_identified_glycopeptides(self) -> List[GlycanComposition]:
        q = self.query(GlycanComposition).join(
                GlycanCombinationGlycanComposition).join(GlycanCombination).join(
                Glycopeptide,
                Glycopeptide.glycan_combination_id == GlycanCombination.id).join(
                IdentifiedGlycopeptide,
                IdentifiedGlycopeptide.structure_id == Glycopeptide.id).filter(
                IdentifiedGlycopeptide.analysis_id == self.analysis_id)
        gcs = [c for c in q]
        return gcs

    def get_glycopeptide_by_sequence(self, sequence: str) -> Query:
        return self.query(IdentifiedGlycopeptide).join(
            IdentifiedGlycopeptide.structure).filter(
                Glycopeptide.glycopeptide_sequence == sequence)

    def get_glycopeptide_spectrum_matches(self, q_value_threshold: float=0.05) -> Query:
        query = self.query(GlycopeptideSpectrumMatch).join(
            GlycopeptideSpectrumSolutionSet).join(
            GlycopeptideSpectrumCluster).join(
            IdentifiedGlycopeptide).filter(
            GlycopeptideSpectrumMatch.is_best_match,
            # GlycopeptideSpectrumMatch.structure_id == IdentifiedGlycopeptide.structure_id,
            GlycopeptideSpectrumMatch.q_value < q_value_threshold,
            GlycopeptideSpectrumMatch.analysis_id == self.analysis_id)
        return query

    def _get_glycopeptide_spectrum_matches_not_matching_owning_structure(self, q_value_threshold: float=0.05) -> Query:
        GPSM_Glycopeptide = aliased(Glycopeptide)
        mixed = self.query(GlycopeptideSpectrumMatch, IdentifiedGlycopeptide).join(
            GlycopeptideSpectrumMatch.solution_set).join(
                GlycopeptideSpectrumSolutionSet.cluster).join(
                    GlycopeptideSpectrumCluster.owners).join(
                        GPSM_Glycopeptide,
                        GlycopeptideSpectrumMatch.structure_id == GPSM_Glycopeptide.id
                    ).join(
                        IdentifiedGlycopeptide.structure
                    ).filter(
            GlycopeptideSpectrumMatch.is_best_match,
            GlycopeptideSpectrumMatch.q_value < q_value_threshold,
            GPSM_Glycopeptide.glycopeptide_sequence != Glycopeptide.glycopeptide_sequence,
            GPSM_Glycopeptide.peptide_id != Glycopeptide.peptide_id,
            GlycopeptideSpectrumMatch.analysis_id == self.analysis_id
        )
        return mixed

    def __repr__(self):
        template = "{self.__class__.__name__}({self.engine!r}, {self.analysis_name!r})"
        return template.format(self=self)


class AnalysisDestroyer(DatabaseBoundOperation):
    def __init__(self, database_connection, analysis_id):
        DatabaseBoundOperation.__init__(self, database_connection)
        self.analysis_id = analysis_id

    def delete_chromatograms(self):
        self.session.query(Chromatogram).filter(
            Chromatogram.analysis_id == self.analysis_id).delete(synchronize_session=False)
        self.session.flush()

    def delete_chromatogram_solutions(self):
        self.session.query(ChromatogramSolution).filter(
            ChromatogramSolution.analysis_id == self.analysis_id).delete(synchronize_session=False)
        self.session.flush()

    def delete_glycan_composition_chromatograms(self):
        self.session.query(GlycanCompositionChromatogram).filter(
            GlycanCompositionChromatogram.analysis_id == self.analysis_id).delete(synchronize_session=False)
        self.session.flush()

    def delete_unidentified_chromatograms(self):
        self.session.query(UnidentifiedChromatogram).filter(
            UnidentifiedChromatogram.analysis_id == self.analysis_id).delete(synchronize_session=False)
        self.session.flush()

    def delete_ambiguous_glycopeptide_groups(self):
        self.session.query(AmbiguousGlycopeptideGroup).filter(
            AmbiguousGlycopeptideGroup.analysis_id == self.analysis_id).delete(synchronize_session=False)
        self.session.flush()

    def delete_identified_glycopeptides(self):
        self.session.query(IdentifiedGlycopeptide).filter(
            IdentifiedGlycopeptide.analysis_id == self.analysis_id).delete(synchronize_session=False)
        self.session.flush()

    def delete_analysis(self):
        self.session.query(Analysis).filter(Analysis.id == self.analysis_id).delete(synchronize_session=False)

    def run(self):
        self.delete_ambiguous_glycopeptide_groups()
        self.delete_identified_glycopeptides()
        self.delete_glycan_composition_chromatograms()
        self.delete_unidentified_chromatograms()
        self.delete_chromatograms()
        self.delete_analysis()
        self.session.commit()


def flatten(iterable):
    return [y for x in iterable for y in x]


class DatabaseScanDeserializer(ScanDeserializerBase, DatabaseBoundOperation):

    def __init__(self, connection, sample_name=None, sample_run_id=None):

        DatabaseBoundOperation.__init__(self, connection)

        self._sample_run = None
        self._sample_name = sample_name
        self._sample_run_id = sample_run_id
        self._iterator = None
        self._scan_id_to_retention_time_cache = None

    def _intialize_scan_id_to_retention_time_cache(self):
        self._scan_id_to_retention_time_cache = dict(
            self.session.query(MSScan.scan_id, MSScan.scan_time).filter(
                MSScan.sample_run_id == self.sample_run_id))

    def __reduce__(self):
        return self.__class__, (
            self._original_connection, self.sample_name, self.sample_run_id)

    # Sample Run Bound Handle API

    @property
    def sample_run_id(self):
        if self._sample_run_id is None:
            self._retrieve_sample_run()
        return self._sample_run_id

    @property
    def sample_run(self):
        if self._sample_run is None:
            self._retrieve_sample_run()
        return self._sample_run

    @property
    def sample_name(self):
        if self._sample_name is None:
            self._retrieve_sample_run()
        return self._sample_name

    def _retrieve_sample_run(self):
        session = self.session
        if self._sample_name is not None:
            sr = session.query(SampleRun).filter(
                SampleRun.name == self._sample_name).one()
        elif self._sample_run_id is not None:
            sr = session.query(SampleRun).filter(
                SampleRun.id == self._sample_run_id).one()
        else:
            sr = session.query(SampleRun).first()
        self._sample_run = sr
        self._sample_run_id = sr.id
        self._sample_name = sr.name

    # Scan Generator Public API

    def get_scan_by_id(self, scan_id):
        q = self._get_by_scan_id(scan_id)
        if q is None:
            raise KeyError(scan_id)
        mem = q.convert()
        if mem.precursor_information:
            mem.precursor_information.source = self
        return mem

    def convert_scan_id_to_retention_time(self, scan_id):
        if self._scan_id_to_retention_time_cache is None:
            self._intialize_scan_id_to_retention_time_cache()
        try:
            return self._scan_id_to_retention_time_cache[scan_id]
        except KeyError:
            q = self.session.query(MSScan.scan_time).filter(
                MSScan.scan_id == scan_id, MSScan.sample_run_id == self.sample_run_id).scalar()
            self._scan_id_to_retention_time_cache[scan_id] = q
            return q

    def _select_index(self, require_ms1=True):
        indices_q = self.session.query(MSScan.index).filter(
            MSScan.sample_run_id == self.sample_run_id).order_by(MSScan.index.asc())
        if require_ms1:
            indices_q = indices_q.filter(MSScan.ms_level == 1)
        indices = flatten(indices_q.all())
        return indices

    def _iterate_over_index(self, start=0, require_ms1=True):
        indices = self._select_index(require_ms1)
        try:
            i = indices.index(start)
        except ValueError:
            lo = 0
            hi = len(indices)

            while lo != hi:
                mid = (lo + hi) // 2
                x = indices[mid]
                if x == start:
                    i = mid
                    break
                elif lo == (hi - 1):
                    i = mid
                    break
                elif x > start:
                    hi = mid
                else:
                    lo = mid
        items = indices[i:]
        i = 0
        n = len(items)
        while i < n:
            index = items[i]
            scan = self.session.query(MSScan).filter(
                MSScan.index == index,
                MSScan.sample_run_id == self.sample_run_id).one()
            products = [pi.product for pi in scan.product_information]
            yield ScanBunch(scan.convert(), [p.convert() for p in products])
            i += 1

    def __iter__(self):
        return self

    def next(self):
        if self._iterator is None:
            self._iterator = self._iterate_over_index()
        return next(self._iterator)

    def _get_by_scan_id(self, scan_id):
        q = self.session.query(MSScan).filter(
            MSScan.scan_id == scan_id, MSScan.sample_run_id == self.sample_run_id).first()
        return q

    def _get_scan_by_time(self, rt, require_ms1=False):
        times_q = self.session.query(MSScan.scan_time).filter(
            MSScan.sample_run_id == self.sample_run_id).order_by(MSScan.scan_time.asc())
        if require_ms1:
            times_q = times_q.filter(MSScan.ms_level == 1)
        times = flatten(times_q.all())
        try:
            i = times.index(rt)
        except ValueError:
            lo = 0
            hi = len(times)

            while lo != hi:
                mid = (lo + hi) // 2
                x = times[mid]
                if x == rt:
                    i = mid
                    break
                elif lo == (hi - 1):
                    i = mid
                    break
                elif x > rt:
                    hi = mid
                else:
                    lo = mid
        scan = self.session.query(MSScan).filter(
            MSScan.scan_time == times[i],
            MSScan.sample_run_id == self.sample_run_id).one()
        return scan

    def reset(self):
        self._iterator = None

    def get_scan_by_time(self, rt, require_ms1=False):
        q = self._get_scan_by_time(rt, require_ms1)
        mem = q.convert()
        if mem.precursor_information:
            mem.precursor_information.source = self
        return mem

    def _get_scan_by_index(self, index):
        q = self.session.query(MSScan).filter(
            MSScan.index == index, MSScan.sample_run_id == self.sample_run_id).one()
        return q

    def get_scan_by_index(self, index):
        mem = self._get_scan_by_index(index).convert()
        if mem.precursor_information:
            mem.precursor_information.source = self
        return mem

    def _locate_ms1_scan(self, scan):
        while scan.ms_level != 1:
            scan = self._get_scan_by_index(scan.index - 1)
        return scan

    def start_from_scan(self, scan_id=None, rt=None, index=None, require_ms1=True):
        if scan_id is None:
            if rt is not None:
                scan = self._get_scan_by_time(rt)
            elif index is not None:
                scan = self._get_scan_by_index(index)
        else:
            scan = self._get_by_scan_id(scan_id)

        # We must start at an MS1 scan, so backtrack until we reach one
        if require_ms1:
            scan = self._locate_ms1_scan(scan)
        self._iterator = self._iterate_over_index(scan.index)
        return self

    # LC-MS/MS Database API

    def msms_for(self, neutral_mass, mass_error_tolerance=1e-5, start_time=None, end_time=None):
        m = neutral_mass
        w = neutral_mass * mass_error_tolerance
        q = self.session.query(PrecursorInformation).join(
            PrecursorInformation.precursor).filter(
            PrecursorInformation.neutral_mass.between(m - w, m + w),
            PrecursorInformation.sample_run_id == self.sample_run_id).order_by(
            MSScan.scan_time)
        if start_time is not None:
            q = q.filter(MSScan.scan_time >= start_time)
        if end_time is not None:
            q = q.filter(MSScan.scan_time <= end_time)
        return q

    def ms1_peaks_above(self, threshold=1000):
        accumulate = [
            (x[0], x[1].convert(), x[1].id) for x in self.session.query(MSScan.scan_id, DeconvolutedPeak).join(
                DeconvolutedPeak).filter(
                MSScan.ms_level == 1, MSScan.sample_run_id == self.sample_run_id,
                DeconvolutedPeak.neutral_mass > threshold
            ).order_by(MSScan.index).yield_per(1000)]
        return accumulate

    def precursor_information(self):
        prec_info = self.session.query(PrecursorInformation).filter(
            PrecursorInformation.sample_run_id == self.sample_run_id).all()
        return prec_info



'''Auxiliary data structures used to look up records from a database
using the Sequence interface or the Mapping interface for convenience.

Maintainence Note: Mostly unused outside of debugging
'''

from glycresoft.serialize import (
    Protein, Peptide, Glycopeptide)

from glycresoft.structure import LRUCache


class ProteinIndex(object):
    def __init__(self, session, hypothesis_id):
        self.session = session
        self.hypothesis_id = hypothesis_id

    def _get_by_id(self, id):
        return self.session.query(Protein).get(id)

    def _get_by_name(self, name):
        return self.session.query(Protein).filter(
            Protein.hypothesis_id == self.hypothesis_id,
            Protein.name == name).one()

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._get_by_id(key)
        else:
            return self._get_by_name(key)

    def __iter__(self):
        q = self.session.query(Protein).filter(Protein.hypothesis_id == self.hypothesis_id)
        return iter(q)

    def __len__(self):
        return self.session.query(Protein).filter(Protein.hypothesis_id == self.hypothesis_id).count()


class PeptideIndex(object):
    def __init__(self, session, hypothesis_id):
        self.session = session
        self.hypothesis_id = hypothesis_id

    def _get_by_id(self, id):
        return self.session.query(Peptide).get(id)

    def _get_by_sequence(self, modified_peptide_sequence, protein_id):
        return self.session.query(Peptide).filter(
            Peptide.hypothesis_id == self.hypothesis_id,
            Peptide.modified_peptide_sequence == modified_peptide_sequence,
            Peptide.protein_id == protein_id).one()

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._get_by_id(key)
        else:
            return self._get_by_sequence(*key)

    def __iter__(self):
        q = self.session.query(Peptide).filter(Peptide.hypothesis_id == self.hypothesis_id)
        return iter(q)

    def __len__(self):
        return self.session.query(Peptide).filter(Peptide.hypothesis_id == self.hypothesis_id).count()



import itertools

from collections import Counter

import numpy as np

from glypy.structure.fragment import Fragment
from glypy.composition import Composition
from glypy.composition.composition_transform import strip_derivatization
from glypy.structure.glycan_composition import MonosaccharideResidue
from glypy.io.nomenclature.identity import is_a

from glycresoft.structure import FragmentMatchMap
from glycresoft.structure.fragment_match_map import SpectrumGraph
from glycresoft.tandem.spectrum_match import SpectrumMatcherBase

from glycopeptidepy.utils.memoize import memoize


dhex = MonosaccharideResidue.from_iupac_lite("dHex")


@memoize(100000000000)
def is_dhex(residue):
    try:
        return is_a(
            strip_derivatization(residue.clone(
                monosaccharide_type=MonosaccharideResidue)), dhex)
    except TypeError:
        if not isinstance(residue, MonosaccharideResidue):
            return False
        else:
            raise


class SignatureIonScorer(SpectrumMatcherBase):
    def __init__(self, scan, glycan_composition, mass_shift=None, use_oxonium=True):
        super(SignatureIonScorer, self).__init__(scan, glycan_composition, mass_shift)
        self.fragments_searched = 0
        self.fragments_matched = 0
        self.minimum_intensity_threshold = 0.01
        self.use_oxonium = use_oxonium

    def _get_matched_peaks(self):
        peaks = set()
        for peak in self.solution_map.by_peak:
            peaks.add(peak)
        for p1, p2, label in self.spectrum_graph:
            peaks.add(p1)
            peaks.add(p2)
        return peaks

    def percent_matched_intensity(self):
        matched = 0
        total = 0

        for peak in self._get_matched_peaks():
            matched += peak.intensity

        for peak in self.spectrum:
            total += peak.intensity
        return matched / total

    def _find_peak_pairs(self, error_tolerance=2e-5, include_compound=False, *args, **kwargs):
        peak_set = self.spectrum
        pairs = SpectrumGraph()

        blocks = [(part, part.mass()) for part in self.target if not is_dhex(part)]
        if include_compound:
            compound_blocks = list(itertools.combinations(self.target, 2))
            compound_blocks = [(block, sum(part.mass() for part in block))
                               for block in compound_blocks]
            blocks.extend(compound_blocks)
        try:
            max_peak = max([p.intensity for p in peak_set])
            threshold = max_peak * self.minimum_intensity_threshold
        except ValueError:
            return []

        for peak in peak_set:
            if peak.intensity < threshold or peak.neutral_mass < 150:
                continue
            for block, mass in blocks:
                for other in peak_set.all_peaks_for(peak.neutral_mass + mass, error_tolerance):
                    if other.intensity < threshold:
                        continue
                    pairs.add(peak, other, block)
        return pairs

    def match(self, error_tolerance=2e-5, include_compound=False, combination_size=3, *args, **kwargs):
        glycan_composition = self.target
        peak_set = self.spectrum
        matches = FragmentMatchMap()
        water = Composition("H2O")
        counter = 0
        try:
            max_peak = max([p.intensity for p in peak_set])
            threshold = max_peak * self.minimum_intensity_threshold
        except ValueError:
            self.solution_map = matches
            self.fragments_searched = counter
            self.pairs = SpectrumGraph()
            return matches
        # Simple oxonium ions
        for k in glycan_composition.keys():
            # dhex does not produce a reliable oxonium ion
            if is_dhex(k):
                continue
            counter += 1
            f = Fragment('B', {}, [], k.mass(), name=str(k),
                         composition=k.total_composition())
            for hit in peak_set.all_peaks_for(f.mass, error_tolerance):
                if hit.intensity < threshold:
                    continue
                matches.add(hit, f)
            f = Fragment('B', {}, [], k.mass() - water.mass, name="%s-H2O" % str(k),
                         composition=k.total_composition() - water)
            for hit in peak_set.all_peaks_for(f.mass, error_tolerance):
                if hit.intensity / max_peak < self.minimum_intensity_threshold:
                    continue
                matches.add(hit, f)

        # Compound oxonium ions
        if include_compound:
            for i in range(2, combination_size + 1):
                for kk in itertools.combinations_with_replacement(sorted(glycan_composition, key=str), i):
                    counter += 1
                    invalid = False
                    for k, v in Counter(kk).items():
                        if glycan_composition[k] < v:
                            invalid = True
                            break
                    if invalid:
                        continue
                    key = '-'.join(map(str, kk))
                    mass = sum(k.mass() for k in kk)
                    composition = sum((k.total_composition() for k in kk), Composition())
                    f = Fragment('B', {}, [], mass, name=key,
                                 composition=composition)
                    for hit in peak_set.all_peaks_for(f.mass, error_tolerance):
                        if hit.intensity / max_peak < 0.01:
                            continue
                        matches.add(hit, f)

                    f = Fragment('B', {}, [], mass - water.mass, name="%s-H2O" % key,
                                 composition=composition - water)
                    for hit in peak_set.all_peaks_for(f.mass, error_tolerance):
                        if hit.intensity / max_peak < 0.01:
                            continue
                        matches.add(hit, f)

        self.spectrum_graph = self._find_peak_pairs(error_tolerance, include_compound)
        self.solution_map = matches
        self.fragments_searched = counter
        return matches

    def penalize(self, ratio, count):
        count = float(count)
        scale = min(np.log(count) / np.log(4), 1)
        return ratio * scale

    def oxonium_ratio(self):
        imax = max(self.spectrum, key=lambda x: x.intensity).intensity
        oxonium = 0
        n = 0
        for peak, fragment in self.solution_map:
            oxonium += peak.intensity / imax
            n += 1
        self.fragments_matched = n
        if n == 0:
            return 0, 0
        return (oxonium / n), n

    def _score_pairs(self):
        try:
            imax = max(self.spectrum, key=lambda x: x.intensity).intensity
        except ValueError:
            return 0
        edge_weight = 0
        n = 0
        for p1, p2, label in self.spectrum_graph:
            edge_weight += (p1.intensity + p2.intensity) / imax
            n += 1
        if n == 0:
            return 0
        scale = max(min(np.log(n) / np.log(4), 1), 0.01)
        return (edge_weight / n) * scale

    def calculate_score(self, error_tolerance=2e-5, include_compound=False, *args, **kwargs):
        if len(self.spectrum) == 0:
            self._score = 0
            return self._score

        use_oxonium = kwargs.get("use_oxonium")
        if use_oxonium is not None:
            self.use_oxonium = use_oxonium

        if self.use_oxonium:
            try:
                oxonium_ratio, n = self.oxonium_ratio()
            except ValueError:
                oxonium_ratio, n = 0, 0
        else:
            oxonium_ratio = n = 0
        if n == 0:
            self._score = 0
        else:
            # simple glycan compositions like high mannose N-glycans don't produce
            # many distinct oxonium ions
            if len(self.target.items()) > 2:
                self._score = self.penalize(oxonium_ratio, n)
            else:
                self._score = oxonium_ratio
        edge_weight = self._score_pairs()
        if edge_weight > self._score:
            self._score = edge_weight
        return self._score



from glycresoft.structure import KeyTransformingDecoratorDict


def key_transform(name):
    return str(name).lower().replace(" ", '-')


hypothesis_register = KeyTransformingDecoratorDict(key_transform)


class BuildBase(object):
    def get_hypothesis_metadata(self):
        raise NotImplementedError()

    @property
    def hypothesis_metadata(self):
        return self.get_hypothesis_metadata()

    def build(self, database_connection, **kwargs):
        raise NotImplementedError()

    def __call__(self, database_connection, **kwargs):
        return self.build(database_connection, **kwargs)



cimport cython

from glypy.composition.ccomposition cimport CComposition


cdef dict mass_shift_index = dict()


cdef class MassShiftBase(object):

    cdef:
        public CComposition composition
        public CComposition tandem_composition
        public basestring name
        public long _hash
        public double mass
        public double tandem_mass


cdef class MassShift(MassShiftBase):
    pass


cdef class CompoundMassShift(MassShiftBase):

    cdef:
        public object counts



cdef MassShiftBase Unmodified



import time
from collections import defaultdict

import numpy as np

from glycresoft.task import TaskBase

from glycresoft.chromatogram_tree import (
    Unmodified,
    ChromatogramFilter,
    ChromatogramOverlapSmoother,
    prune_bad_mass_shift_branches)

from glycresoft.scoring import (
    ChromatogramSolution,
    ChromatogramScorer)

from glycresoft.composition_distribution_model import smooth_network, display_table


class ChromatogramEvaluator(TaskBase):
    acceptance_threshold = 0.4
    ignore_below = 1e-5

    def __init__(self, scoring_model=None):
        if scoring_model is None:
            scoring_model = ChromatogramScorer()
        self.scoring_model = scoring_model

    def configure(self, analysis_info):
        self.scoring_model.configure(analysis_info)

    def evaluate(self, chromatograms, delta_rt=0.25, min_points=3, smooth_overlap_rt=True,
                 *args, **kwargs):
        filtered = ChromatogramFilter.process(
            chromatograms, delta_rt=delta_rt, min_points=min_points)
        if smooth_overlap_rt:
            filtered = ChromatogramOverlapSmoother(filtered)
        solutions = []
        i = 0
        n = len(filtered)
        for case in filtered:
            start = time.time()
            i += 1
            if self.in_debug_mode():
                self.debug("... Evaluating %r" % (case, ))
            if i % 1000 == 0:
                self.log("... %0.2f%% chromatograms evaluated (%d/%d)" % (i * 100. / n, i, n))
            try:
                sol = self.evaluate_chromatogram(case)
                if self.scoring_model.accept(sol):
                    solutions.append(sol)
                else:
                    if sol.glycan_composition:
                        self.debug("... Rejecting %s with score %s %s" % (
                            sol, sol.score, sol.score_components()))
                end = time.time()
                # Report on anything that took more than 30 seconds to evaluate
                if end - start > 30.0:
                    self.log("... %r took a long time to evaluated (%0.2fs)" % (case, end - start))
            except (IndexError, ValueError):
                continue
        solutions = ChromatogramFilter(solutions)
        return solutions

    def evaluate_chromatogram(self, chromatogram):
        score_set = self.scoring_model.compute_scores(chromatogram)
        score = score_set.product()
        return ChromatogramSolution(
            chromatogram, score, scorer=self.scoring_model,
            score_set=score_set)

    def finalize_matches(self, solutions):
        out = []
        for sol in solutions:
            if sol.score <= self.ignore_below:
                continue
            elif (sol.composition is None) and (Unmodified not in sol.mass_shifts):
                continue
            out.append(sol)
        solutions = ChromatogramFilter(out)
        return solutions

    def score(self, chromatograms, delta_rt=0.25, min_points=3, smooth_overlap_rt=True,
              mass_shifts=None, *args, **kwargs):

        solutions = self.evaluate(
            chromatograms, delta_rt, min_points, smooth_overlap_rt, *args, **kwargs)

        if mass_shifts:
            self.log("Pruning mass shift branches")
            hold = self.prune_mass_shifts(solutions)
            self.log("Re-evaluating after mass shift pruning")
            solutions = self.evaluate(hold, delta_rt, min_points, smooth_overlap_rt,
                                      *args, **kwargs)

        solutions = self.finalize_matches(solutions)
        return solutions

    def prune_mass_shifts(self, solutions):
        return prune_bad_mass_shift_branches(ChromatogramFilter(solutions))

    def acceptance_filter(self, solutions, threshold=None):
        if threshold is None:
            threshold = self.acceptance_threshold
        return ChromatogramFilter([
            sol for sol in solutions
            if sol.score >= threshold and not sol.used_as_mass_shift
        ])

    def update_parameters(self, param_dict):
        param_dict['scoring_model'] = self.scoring_model


class LogitSumChromatogramEvaluator(ChromatogramEvaluator):
    acceptance_threshold = 4
    ignore_below = 2
    update_score_on_merge = True

    def __init__(self, scoring_model):
        super(LogitSumChromatogramEvaluator, self).__init__(scoring_model)

    def prune_mass_shifts(self, solutions):
        return prune_bad_mass_shift_branches(ChromatogramFilter(solutions), score_margin=2.5)

    def evaluate_chromatogram(self, chromatogram):
        score_set = self.scoring_model.compute_scores(chromatogram)
        logitsum_score = score_set.logitsum()
        return ChromatogramSolution(
            chromatogram, logitsum_score, scorer=self.scoring_model,
            score_set=score_set)

    def evaluate(self, chromatograms, delta_rt=0.25, min_points=3, smooth_overlap_rt=True,
                 *args, **kwargs):
        solutions = super(LogitSumChromatogramEvaluator, self).evaluate(
            chromatograms, delta_rt=delta_rt, min_points=min_points,
            smooth_overlap_rt=smooth_overlap_rt, *args, **kwargs)
        self.log("Collapsing Duplicates")
        accumulator = defaultdict(list)
        for case in solutions:
            accumulator[case.key].append(case)
        solutions = []
        n = len(accumulator)
        i = 0.0
        for group, members in accumulator.items():
            if i % 1000 == 0 and i > 0:
                self.log("... %d groups collapsed (%0.02f%%)" % (i, i / n * 100.0))
            members = sorted(members, key=lambda x: x.score, reverse=True)
            reference = members[0]
            base = reference.clone()
            for other in members[1:]:
                base = base.merge(other, skip_duplicate_nodes=True)
            merged = reference.__class__(
                base, reference.score, scorer=reference.scorer,
                score_set=reference.score_set)
            if self.update_score_on_merge and len(members) > 1:
                aggregated = self.evaluate_chromatogram(merged)
                if aggregated.score > reference.score:
                    merged.score_set = aggregated.score_set
                    merged.score = aggregated.score
            solutions.append(merged)
            i += 1.0
        return ChromatogramFilter(solutions)


class LaplacianRegularizedChromatogramEvaluator(LogitSumChromatogramEvaluator):
    def __init__(self, scoring_model, network, smoothing_factor=None, grid_smoothing_max=1.0,
                 regularization_model=None):
        super(LaplacianRegularizedChromatogramEvaluator,
              self).__init__(scoring_model)
        self.network = network
        self.smoothing_factor = smoothing_factor
        self.grid_smoothing_max = grid_smoothing_max
        self.regularization_model = regularization_model

    def network_smoothing(self, solutions):
        is_bipartite = False
        if isinstance(self.smoothing_factor, tuple):
            smoothing_factor = self.smoothing_factor[0]
            is_bipartite = True
        else:
            smoothing_factor = self.smoothing_factor
        updated_network, search, params = smooth_network(
            self.network, solutions, lmbda=smoothing_factor,
            lambda_max=self.grid_smoothing_max,
            model_state=self.regularization_model)
        if is_bipartite:
            regularization_model = params
            updated_network, search, params = smooth_network(
                self.network, solutions, lmbda=self.smoothing_factor[1],
                lambda_max=self.grid_smoothing_max,
                model_state=regularization_model)
        return updated_network, search, params

    def evaluate(self, chromatograms, delta_rt=0.25, min_points=3, smooth_overlap_rt=True,
                 *args, **kwargs):
        solutions = super(LaplacianRegularizedChromatogramEvaluator, self).evaluate(
            chromatograms, delta_rt=delta_rt, min_points=min_points,
            smooth_overlap_rt=smooth_overlap_rt, *args, **kwargs)
        self.log("... Applying Network Smoothing Regularization")
        updated_network, search, params = self.network_smoothing(solutions)
        if updated_network is search is params is None:
            return solutions
        solutions = sorted(solutions, key=lambda x: x.score, reverse=True)
        # TODO - Use aggregation across multiple observations for the same glycan composition
        # instead of discarding all but the top scoring feature?
        seen = dict()
        unannotated = []
        for sol in solutions:
            if sol.glycan_composition is None:
                unannotated.append(sol)
                continue
            if sol.glycan_composition in seen:
                continue
            seen[sol.glycan_composition] = sol
            node = updated_network[sol.glycan_composition]
            if sol.score > self.acceptance_threshold:
                sol.score = node.score
            else:
                # Do not permit network smoothing to boost scores below acceptance_threshold
                if node.score < sol.score:
                    sol.score = node.score
        self.network_parameters = params
        self.grid_search = search
        display_table(
            search.model.neighborhood_names,
            np.array(params.tau).reshape((-1, 1)),
            print_fn=lambda x: self.log("...... %s" % (x,)))
        self.log("...... smoothing factor: %0.3f; threshold: %0.3f" % (
            params.lmbda, params.threshold))
        return ChromatogramFilter(list(seen.values()) + unannotated)

    def update_parameters(self, param_dict):
        super(LaplacianRegularizedChromatogramEvaluator, self).update_parameters(param_dict)
        param_dict['network_parameters'] = self.network_parameters
        param_dict['network_model'] = self.grid_search



from glypy.composition import formula

from .proteomics.peptide_permutation import peptide_permutations
from glycresoft.serialize import Peptide, DatabaseBoundOperation


class ProteomeCollection(object):
    def __init__(self, proteins, peptides):
        self.proteins = proteins
        self.peptides = peptides

    def add_modifications(self, constant_modifications=None, variable_modifications=None, max_variable_modifications=4):
        if constant_modifications is None:
            constant_modifications = []
        if variable_modifications is None:
            variable_modifications = []
        result = MemoryPeptideCollection()
        for peptide in self.peptides:
            for modified_peptide, n_variable_modifications in peptide_permutations(
                    str(peptide), constant_modifications, variable_modifications):
                total_modification_count = (
                    n_variable_modifications + peptide.count_variable_modifications)
                if total_modification_count > max_variable_modifications:
                    continue
                inst = Peptide(
                    base_peptide_sequence=peptide.base_peptide_sequence,
                    modified_peptide_sequence=str(modified_peptide),
                    count_missed_cleavages=peptide.count_missed_cleavages,
                    count_variable_modifications=total_modification_count,
                    sequence_length=peptide.sequence_length,
                    start_position=peptide.start_position,
                    end_position=peptide.end_position,
                    calculated_mass=modified_peptide.mass,
                    formula=formula(modified_peptide.total_composition()))
                result.add(inst)
        return result


class PeptideCollectionBase(object):
    def __init__(self, *args, **kwargs):
        pass

    def __iter__(self):
        raise NotImplementedError()

    def add(self, peptide):
        raise NotImplementedError()

    def update(self, peptides):
        for peptide in peptides:
            self.add(peptide)

    def save(self):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def __repr__(self):
        return "{self.__class__.__name__}({size})".format(self=self, size=len(self))


class MemoryPeptideCollection(PeptideCollectionBase):
    def __init__(self, storage=None, *args, **kwargs):
        PeptideCollectionBase.__init__(self, *args, **kwargs)
        self.storage = list(storage or [])

    def __iter__(self):
        return iter(self.storage)

    def __len__(self):
        return len(self.storage)

    def add(self, peptide):
        self.storage.append(peptide)

    def save(self):
        pass


class DatabasePeptideCollection(PeptideCollectionBase, DatabaseBoundOperation):
    def __init__(self, connection, hypothesis_id=None, *args, **kwargs):
        if hypothesis_id is None:
            hypothesis_id = 1
        DatabaseBoundOperation.__init__(self, connection)
        PeptideCollectionBase.__init__(self, *args, **kwargs)
        self.hypothesis_id = hypothesis_id
        self._operation_count = 0
        self._batch_size = int(kwargs.get("batch_size", 1000))

    def __iter__(self):
        q = self._get_query()
        q = q.yield_per(5000)
        return q

    def _get_query(self):
        q = self.session.query(Peptide).filter(Peptide.hypothesis_id == self.hypothesis_id)
        return q

    def __len__(self):
        return self._get_query().count()

    def _add(self, peptide):
        peptide.hypothesis_id = self.hypothesis_id
        self.session.add(peptide)
        self._operation_count += 1

    def add(self, peptide):
        self._add(peptide)
        self._flush_if_work_done()

    def update(self, peptides):
        for peptide in peptides:
            self._add(peptide)
        self._flush_if_work_done()

    def _flush_if_work_done(self):
        if self._operation_count > self._batch_size:
            self.session.flush()

    def save(self):
        self.session.commit()



import unittest
import tempfile
import os


from . import fixtures
from .test_constrained_combinatorics import (
    FILE_SOURCE)
from glycresoft.profiler import MzMLGlycanChromatogramAnalyzer, GeneralScorer
from glycresoft.database.builder.glycan import CombinatorialGlycanHypothesisSerializer
from glycresoft.serialize import AnalysisDeserializer


agp_glycomics_mzml = fixtures.get_test_data(
    "AGP_Glycomics_20150930_06.deconvoluted.mzML")


class GlycanProfilerConsumerTest(unittest.TestCase):

    def setup_tempfile(self, content):
        file_name = tempfile.mktemp() + '.tmp'
        open(file_name, 'w').write(content)
        return file_name

    def clear_file(self, path):
        open(path, 'wb').close()

    def _make_hypothesis(self):
        file_name = self.setup_tempfile(FILE_SOURCE)
        builder = CombinatorialGlycanHypothesisSerializer(
            file_name, file_name + '.db')
        builder.run()
        self.clear_file(file_name)
        return file_name + '.db'

    def confirm_score(self, chroma, key, score):
        match = chroma.find_key(key)
        self.assertIsNotNone(match)
        self.assertAlmostEqual(score, match.score, 1)

    def confirm_absent(self, chroma, key):
        match = chroma.find_key(key)
        self.assertIsNone(match)

    def test_profiler(self):
        db_file = self._make_hypothesis()
        output_file = self.setup_tempfile("")
        task = MzMLGlycanChromatogramAnalyzer(
            db_file, 1, agp_glycomics_mzml, output_file,
            analysis_name="test-analysis",
            scoring_model=GeneralScorer)
        task.start()
        self.assertTrue(os.path.exists(output_file))
        ads = AnalysisDeserializer(output_file)
        gcs = ads.load_glycan_composition_chromatograms()
        self.assertEqual(len(gcs), 23)
        self.clear_file(db_file)
        # 'spacing_fit': 0.96367957815527916, 'isotopic_fit': 0.99366937970680247,
        # 'line_score': 0.99780414736388745, 'charge_count': 0.9365769766604084
        self.confirm_score(gcs, "{Fuc:1; Hex:7; HexNAc:6; Neu5Ac:4}", 17.1458)
        # 'spacing_fit': 0.96123524755239487, 'isotopic_fit': 0.97935840584492162,
        # 'line_score': 0.99562733579066764, 'charge_count': 0.7368321292716115
        self.confirm_score(gcs, "{Hex:8; HexNAc:7; Neu5Ac:3}", 13.5279)
        # 'spacing_fit': 0.94565181061625481, 'isotopic_fit': 0.99074210231338733,
        # 'line_score': 0.98925755528448378, 'charge_count': 0.999773289306269
        self.confirm_score(gcs, "{Hex:7; HexNAc:6; Neu5Ac:4}", 20.4438)
        # 'spacing_fit': 0.95567017048597336, 'isotopic_fit': 0.98274665306540443,
        # 'line_score': 0.99706887771172914, 'charge_count': 0.7604540961453831
        self.confirm_score(gcs, "{Fuc:2; Hex:6; HexNAc:5; Neu5Ac:3}", 14.0977)

        ads.close()
        self.clear_file(output_file)

    def test_smoothing_profiler(self):
        db_file = self._make_hypothesis()
        output_file = self.setup_tempfile("")
        task = MzMLGlycanChromatogramAnalyzer(
            db_file, 1, agp_glycomics_mzml, output_file,
            regularize="grid",
            analysis_name="test-analysis",
            scoring_model=GeneralScorer)
        task.start()
        # import cProfile
        # prof = cProfile.Profile()
        # prof.runcall(task.start)
        # prof.print_stats()
        # prof.dump_stats('smooth_profile.pstats')
        self.assertTrue(os.path.exists(output_file))
        ads = AnalysisDeserializer(output_file, analysis_id=1)
        gcs = ads.load_glycan_composition_chromatograms()
        self.assertEqual(len(gcs), 23)
        self.confirm_score(gcs, "{Fuc:1; Hex:7; HexNAc:6; Neu5Ac:4}", 16.1425)
        self.confirm_score(gcs, "{Hex:8; HexNAc:7; Neu5Ac:3}", 8.8510)
        self.confirm_score(gcs, "{Hex:7; HexNAc:6; Neu5Ac:4}", 16.6722)
        network_params = ads.analysis.parameters['network_parameters']
        tau = [0.0, 12.173488161057854, 16.042106463675424, 0.0, 22.061954223206591,
               0.0, 13.928596053020485, 0.0, 9.4348332520855713, 0.0, 0.0, 0.0, 0.0, 0.0]
        for a, b in zip(tau, network_params.tau):
            self.assertAlmostEqual(a, b, 3)
        ads.close()

        self.clear_file(output_file)
        task = MzMLGlycanChromatogramAnalyzer(
            db_file, 1, agp_glycomics_mzml, output_file,
            regularize=0.2,
            regularization_model=network_params,
            analysis_name="test-analysis",
            scoring_model=GeneralScorer)
        task.start()
        ads = AnalysisDeserializer(output_file, analysis_id=1)
        gcs = ads.load_glycan_composition_chromatograms()
        self.assertEqual(len(gcs), 23)
        self.confirm_score(gcs, "{Fuc:1; Hex:7; HexNAc:6; Neu5Ac:4}", 16.7795)
        self.confirm_score(gcs, "{Hex:8; HexNAc:7; Neu5Ac:3}", 10.6734)
        self.confirm_score(gcs, "{Hex:7; HexNAc:6; Neu5Ac:4}", 18.3360)
        self.confirm_score(gcs, "{Fuc:2; Hex:6; HexNAc:5; Neu5Ac:3}", 15.9628)
        network_params = ads.analysis.parameters['network_parameters']
        for a, b in zip(tau, network_params.tau):
            self.assertAlmostEqual(a, b, 3)
        ads.close()
        self.clear_file(output_file)
        self.clear_file(db_file)


if __name__ == '__main__':
    unittest.main()
